from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np
from PyAstronomy.pyasl import _ic


def equiTempPlanet(Ts, R, a, albedo=0.0, e=0.0, solarUnits=False):
  """
    Calculate the planetary equilibrium temperature.
    
    The equilibrium temperature is defined, e.g., in
    Laughlin et al. 2011 (ApJ, 729), Eq. 1. Albedo
    was added "by hand".
    
    Note that units have to be consistent for stellar
    radius and large semi-major axis. The `solarUnits`
    flag can be used to use solar units for these
    quantities.
    
    Parameters
    ----------
    Ts : float
        Stellar effective temperature [K].
    R : float
        Radius of the star
    a : float
        Large semi-major axis
    albedo : float
        Albedo of the planetary atmosphere.
        Default is 0.0.
    e : float, optional
        Eccentricity of orbit (default is 0.0).
    solarUnits : boolean, optional
        If this flag is set True, stellar radius and
        semi-major axis are taken to be in solar radii
        and AU.
    
    Returns
    -------
    Equilibrium temperature : float
        The equilibrium temperature of the planet.
  """
  if _ic.check["quantities"]:
    from PyAstronomy import constants as c
  else:
    raise(PE.PyARequiredImport("You need to install 'quantities' to use this function."))
    return None
  if solarUnits:
    a = a * c.AU
    R = R * c.RSun
  if a < R:
    raise(PE.PyAValError("The semi-major axis cannot be smaller than the radius of the star."))
  teq = np.power(1. - albedo, 1./4.) * np.sqrt(R / (2.0 * a)) * Ts \
        / np.power(1.0 - e**2, 1./8.)
  return teq
  


from __future__ import print_function
import unittest
import os


class TestSanityOfmtools(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        if os.path.isfile("test.tmp"):
            os.remove("test.tmp")

    def testsanity_numericalDerivative(self):
        """
        mtools: Checking accuracy of numerical derivatives.
        """
        # Check polynomial
        from .numericalDerivatives import diffCFD
        import numpy as np

        # Analytical derivatives
        x = np.linspace(-10, 10, 1000)
        y = [np.poly1d((0.03, -0.31, 0.4, 0.35, 1.4))]
        for i in range(4):
            y.append(y[-1].deriv())

        for erro in [2, 4, 6]:
            for i in range(1, 5):
                indi, der = diffCFD(x, np.polyval(y[0], x), i, erro)
                self.assertLess(
                    np.max(np.abs(der / np.polyval(y[i], x[indi]) - 1.0)),
                    2e-2 * pow(10.0, erro - 2),
                )

        # Check trigonometric
        y = [np.sin(x / 10.0 * 2 * np.pi + 3)]
        y.append(2 * np.pi / 10.0 * np.cos(x / 10.0 * 2 * np.pi + 3))
        y.append(-((2 * np.pi / 10.0) ** 2) * np.sin(x / 10.0 * 2 * np.pi + 3))
        y.append(-((2 * np.pi / 10.0) ** 3) * np.cos(x / 10.0 * 2 * np.pi + 3))
        y.append((2 * np.pi / 10.0) ** 4 * np.sin(x / 10.0 * 2 * np.pi + 3))
        for erro in [2, 4, 6]:
            for i in range(1, 5):
                indi, der = diffCFD(x, y[0], i, erro)
                self.assertLess(
                    np.max(np.abs(der / y[i][indi] - 1.0)), 1e-3 * pow(10.0, erro - 2)
                )

        # Check exponential
        y = np.exp(x)
        for erro in [2, 4, 6]:
            for i in range(1, 5):
                print(i, erro)
                indi, der = diffCFD(x, y, i, erro)
                self.assertLess(
                    np.max(np.abs(der / y[indi] - 1.0)), 1e-3 * pow(10.0, erro - 2)
                )

    def testsanity_diffCFDExample(self):
        """
        mtools: diffCFD example
        """
        from PyAstronomy import pyaC
        import matplotlib.pylab as plt
        import numpy as np

        x = np.linspace(-10, 10, 1000)

        # Computer polynomial and its derivatives
        # (quasi analytically)
        y = [np.poly1d((0.03, -0.31, 0.4, 0.35, 1.4))]
        for i in range(4):
            y.append(y[-1].deriv())

        # Compute derivates numerically and compare to
        # analytic solution
        erro = 2
        for i in range(1, 5):
            indi, der = pyaC.diffCFD(x, np.polyval(y[0], x), i, erro)
            plt.plot(x[indi], np.polyval(y[i], x[indi]), "b.")
            plt.plot(x[indi], der, "r--")

    #      plt.show()

    def testsanity_ibtrapzExample(self):
        """
        mtools: Checking example of ibtrapz
        """
        from PyAstronomy.pyaC import mtools
        import numpy as np

        x = np.arange(-2.0, 2.01, 0.1)
        y = x**3 + 1.7

        x0 = -1.375
        x1 = +1.943

        # Analytical value of integral
        analyt = 0.25 * (x1**4 - x0**4) + 1.7 * (x1 - x0)

        print("Analytical value: ", analyt)
        print("ibtrapz: ", mtools.ibtrapz(x, y, x0, x1))

    def testsanity_ibtrapz(self):
        """
        mtools: Checking ibtrapz
        """
        from PyAstronomy.pyaC import mtools
        import numpy as np

        x = np.arange(-2.0, 2.01, 0.1)
        y = 2.0 * x

        x0 = -1.375
        x1 = +1.943

        # Analytical value of integral
        analyt = x1**2 - x0**2

        self.assertAlmostEqual(
            analyt,
            mtools.ibtrapz(x, y, x0, x1),
            delta=1e-10,
            msg="ibtrapz incorrect for linear function.",
        )
        self.assertAlmostEqual(
            (-1.9) ** 2 - (-2.0) ** 2,
            mtools.ibtrapz(x, y, -2.0, -2.0 + 0.1),
            delta=1e-10,
            msg="ibtrapz incorrect for linear function (-2,-1.9).",
        )
        self.assertAlmostEqual(
            0.0,
            mtools.ibtrapz(x, y, -2.0, +2.0),
            delta=1e-10,
            msg="ibtrapz incorrect for linear function (-2,+2).",
        )

    def testsanity_zerocross1dExample(self):
        """
        Checking sanity of zerocross1d example
        """
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import pyaC

        # Generate some 'data'
        x = np.arange(100.0) ** 2
        y = np.sin(x)

        # Set the last data point to zero.
        # It will not be counted as a zero crossing!
        y[-1] = 0

        # Set point to zero. This will be counted as a
        # zero crossing
        y[10] = 0.0

        # Get coordinates and indices of zero crossings
        xc, xi = pyaC.zerocross1d(x, y, getIndices=True)

        # Plot the data
        plt.plot(x, y, "b.-")
        # Add black points where the zero line is crossed
        plt.plot(xc, np.zeros(len(xc)), "kp")
        # Add green points at data points preceding an actual
        # zero crossing.
        plt.plot(x[xi], y[xi], "gp")

    #     plt.show()

    def testsanity_zerocross1d(self):
        """
        Checking sanity of zerocross1d
        """
        import numpy as np
        from PyAstronomy import pyaC

        x = np.arange(3.0)
        y = np.array([0, 3, 0])

        xz = pyaC.zerocross1d(x, y)
        self.assertEqual(
            len(xz),
            0,
            msg="Found zero crossing in 0,3,0 array (problem with first/last point).",
        )

        y = np.array([-1.0, 1.0, -2.0])
        xz, xi = pyaC.zerocross1d(x, y, getIndices=True)
        self.assertEqual(
            len(xz),
            2,
            msg="Found the following zero crossings in -1,1,-2 array: " + str(xz),
        )
        self.assertEqual(
            len(xz), len(xi), "Number of values and indicies is not identical."
        )

        self.assertAlmostEqual(
            np.max(np.abs(np.array([0.5, 1.0 + 1.0 / 3.0]) - xz)),
            0.0,
            delta=1e-8,
            msg="Found unexpected zero crossings: " + str(xz),
        )
        self.assertEqual(
            np.max(np.abs(xi - np.array([0, 1]))),
            0,
            msg="Found unexpected indices: " + str(xi),
        )

    def testsanity_farat(self):
        """
        Checking sanity of farat
        """
        from PyAstronomy import pyaC

        x = pyaC.farat(5, 3)
        y = pyaC.farat(3, 5)
        self.assertAlmostEqual(
            x, 1.0 / y, delta=1e-10, msg="farat(5,3) and farat(3,5) not consistent."
        )
        self.assertAlmostEqual(
            x,
            20.0,
            delta=1e-10,
            msg="farat(5,3) broken (returned value of " + str(x) + ")",
        )



from __future__ import print_function
from .PyA_Version import PyA_Version
import os
import importlib
from PyAstronomy.pyaC import pyaPermanent

class PyAIS:
    """
    PyAstronomy Import Status
    """

    def findpackages(self):
        for root, dirs, fns in os.walk(__path__[0]):
            for dir in dirs:
                if os.path.isfile(os.path.join(__path__[0], dir, "__init__.py")):
                    self.packages.append(dir)
            break

    def showStatus(self):
        print()
        print("        Results of PyAIS import check:")
        print("-----------------------------------------------------")
        print("%20s  |  %5s" % ("Package Name", "Import Status"))
        print("-----------------------------------------------------")
        for m, v in sorted(self.importStatus.items()):
            print("%20s  |  %5s" % (m, v))

    def __init__(self):
        """
        Try to import all packages contained in self.packages and save status.
        """

        self.packages = []
        self.findpackages()

        self.importStatus = {}
        for m in self.packages:
            try:
                importlib.import_module("." + m, "PyAstronomy")
                self.importStatus[m] = True
            except Exception:
                self.importStatus[m] = False


def importCheck():
    p = PyAIS()
    p.showStatus()

pyaRC = pyaPermanent.pyaRC

# Check for the presence of six
try:
    import six
except ImportError:
    raise(ImportError("""
  
  The 'six' package could not be imported.
  
  Please install 'six' to use PyAstronomy. The package
  helps a great deal in providing Python 2/3 (2*3=six) compatible
  code.
  
  """))


# The version Tag for PyAstronomy
__version__ = PyA_Version()



import unittest


class TestForTransExampletestsanity_(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity__calcModel_1(self):
        # Import some unrelated modules
        import numpy as np
        import matplotlib.pylab as plt

        # ... and now the forTrans module
        from PyAstronomy.modelSuite import forTrans as ft

        # Create MandelAgolLC object with
        # circular orbit and quadratic limb darkening
        ma = ft.MandelAgolLC(orbit="circular", ld="quad")

        # See the available parameters and their current values
        ma.parameterSummary()

        # Set parameters
        ma["per"] = 0.2
        ma["i"] = 90.0
        ma["a"] = 6.5
        ma["T0"] = 0.0
        ma["p"] = 0.16
        ma["linLimb"] = 0.47
        ma["quadLimb"] = 0.24
        ma["b"] = 0.0

        # Choose some time axis
        time = np.linspace(0, 0.5, 1000)

        # ... and calculate model
        y = ma.evaluate(time)

        # Let's see what happened ...
        plt.plot(time, y, "b.")
        # plt.show()

    def testsanity__calcModel_2(self):
        # Import some unrelated modules
        import numpy as np
        import matplotlib.pylab as plt

        # ... and now the forTrans module
        from PyAstronomy.modelSuite import forTrans as ft

        # Create MandelAgolLC object with
        # keplerian orbit and quadratic limb darkening
        ma = ft.MandelAgolLC(orbit="keplerian", ld="quad")

        # See the available parameters and their current values
        ma.parameterSummary()

        # Set parameters
        ma["per"] = 0.2
        ma["i"] = 88.76
        ma["a"] = 6.5
        ma["p"] = 0.16
        ma["linLimb"] = 0.47
        ma["quadLimb"] = 0.24
        ma["b"] = 0.0
        ma["e"] = 0.75
        ma["w"] = 127.0
        ma["Omega"] = 3.9

        # Choose some time axis
        time = np.linspace(0, 0.5, 1000)

        # ... and calculate model
        y = ma.evaluate(time)

        # Let's see what happened ...
        plt.plot(time, y, "b.")
        # plt.show()

    def testsanity__CompareLDLaws(self):
        # Import some modules
        import numpy as np
        import matplotlib.pylab as plt

        # ... and now the forTrans module
        from PyAstronomy.modelSuite import forTrans as ft

        # First, let's compute a transit model using
        # quadratic limb-darkening prescription.
        ma = ft.MandelAgolLC(ld="quad")

        # Set parameters. The LD coefficients are taken
        # from Claret 2011 for a solar-metallicity star
        # with Teff=6000 K and logg=4.5.
        ma["per"] = 0.2
        ma["i"] = 90.0
        ma["a"] = 6.5
        ma["T0"] = 0.5
        ma["p"] = 0.16
        ma["linLimb"] = 0.0479
        ma["quadLimb"] = 0.2716
        ma["b"] = 0.0

        # Choose some time axis
        time = np.linspace(0, 0.2, 1000)

        # ... and calculate model
        yQLD = ma.evaluate(time)

        # Now, let's compute a transit model with
        # non-linear limb-darkening prescription
        # for the same stellar parameters.
        maNL = ft.MandelAgolLC(ld="nl")
        maNL["per"] = 0.2
        maNL["i"] = 90.0
        maNL["a"] = 6.5
        maNL["T0"] = 0.5
        maNL["p"] = 0.16
        maNL["a1"] = 0.5335
        maNL["a2"] = 0.0793
        maNL["a3"] = -0.3466
        maNL["a4"] = 0.1609
        maNL["b"] = 0.0

        yNLLD = maNL.evaluate(time)

        # Let's compare both models...
        plt.plot(time, yQLD, "-", label="Quadratic LD")
        plt.plot(time, yNLLD, "d", label="Non-linear LD")
        plt.legend()
        # plt.show()


class TestFortranPyImplementation(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def testsanity_py_vs_fortran(self):
        """
        Check python vs fortran implementation (occultquad)
        """
        import numpy as np
        # ... and now the forTrans module
        from PyAstronomy.modelSuite import forTrans as ft
        
        
        # Create MandelAgolLC object with
        # circular orbit and quadratic limb darkening
        ma = ft.MandelAgolLC(orbit="circular", ld="quad", pyfo=False)
        mapy = ft.MandelAgolLC(orbit="circular", ld="quad", pyfo=True)

        np.random.seed(1819)
        # Choose some time axis
        time = np.linspace(-0.5, 0.5, 2000)
        
        for i in range(100):
        
            # Set parameters
            ma["per"] = 1
            ma["a"] = 4
            ma["p"] = np.random.random()*0.1 + 1e-4
            imin = 90 - np.arcsin( (1+ma["p"])/ma["a"] )/180*np.pi
            ma["i"] = np.random.random() * (89-imin) + imin
            ma["T0"] = 0.0
            ma["linLimb"] = np.random.random()
            ma["quadLimb"] = np.random.random() * (1-ma["linLimb"])
            ma["b"] = 0.
        
            mapy.assignValues(ma.parameters())
            
            y = ma.evaluate(time)
            ypy = mapy.evaluate(time)
            
            maxab = np.max(np.abs(y-ypy))
            
            ma.parameterSummary()
            self.assertAlmostEqual(maxab, 0, 7, msg=f"Problem with python vs fortran implementations at position {i}")
            
    



import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Sampling from linear model
        # =========================
        
        # A dataset consisting of three simultaneously obtained data sets (e.g., different photometric bands or RVs determined in individual spectral orders) is constructed and a model is set up with a single slope for all series but individual means. 
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit2 as fuf2
        import scipy.optimize as sco
        import matplotlib
        matplotlib.use('Agg')
        
        # In[2]:
        
        
        # Definition of model 
        
        class LinMod(fuf2.MBOEv):
            """ Linear model """
        
            def __init__(self, no):
                # no is number of measurement series
                self.no = no
                # 'parnames' specifies parameter names in the model
                parnames = ["mean%d" % (i+1) for i in range(no)]
                parnames.append("slope")
                
                fuf2.MBOEv.__init__(self, pars=parnames, rootName="LinMod")
        
            def evaluate(self, x, *args, **kwargs):
                """
                Evaluate model at 'x' here and return y = const + x*slope
                """
                r = np.zeros( (len(x), self.no) )
                for i in range(self.no):
                    r[::,i] = self["slope"] * x + self["mean%d" % (i+1)]
                
                return r
        
        # In[3]:
        
        
        # Generate mock data
        np.random.seed(1718)
        
        # Number of data points and orders
        nd = 10
        no = 3
        # Uncertainty of individual points
        s = 0.3
        
        # Mock data and time (x) axis
        d = np.zeros((nd,no))
        x = np.arange(nd)
        
        for i in range(no):
            offset = np.random.random() * 20
            d[::,i] = 0.6 * x + offset + np.random.normal(0, s, nd)
        
        # Uncertainty
        yerr = np.ones_like(d)*s
        
        # In[4]:
        
        
        # Get best-fit parameters
        
        # Instantiate model
        lm = LinMod(no)
        # Guess slope
        lm["slope"] = 0.5
        # Thaw parameters
        lm.thaw(["mean%d" % (i+1) for i in range(no)])
        lm.thaw(["slope"])
        
        # Because ln inherited from MBOEv, it has a default chi-square objective function
        # Find best-fit parameters
        fr = sco.fmin(lm.chisqr, x0=lm.freeParamVals(), args=(x,d,yerr))
        # Set model parameters to best-fit values and display
        lm.setFreeParamVals(fr)
        lm.parameterSummary()
        
        # Plot best-fit model
        plt.title("Data and model")
        model = lm.evaluate(x)
        for i in range(no):
            plt.errorbar(x, d[::,i], yerr=s, fmt='+-')
            plt.plot(x, model[::,i], 'k:')
        plt.show()
        
        # In[5]:
        
        
        # Sample from the posterior
        
        fuf2.sampleEMCEE2(lm, pargs=(x, d, yerr), dbfile="chain1.emcee", sampleArgs={"burn":300, "iters":1000})
        
        
        ta = fuf2.TraceAnalysis2("chain1.emcee")
        
        # Calculate mean, median, standard deviation, and
        # credibility interval for the available parameters
        for p in ta.availableParameters():
          hpd = ta.hpd(p, cred=0.95)
          print("Parameter %5s, mean = % g, median = % g, std = % g, 95%% HPD = % g - % g" \
                % (p, ta.mean(p), ta.median(p), ta.std(p), hpd[0], hpd[1]))
        
        ta.plotTraceHist("slope")
        ta.show()
        



from __future__ import print_function
from __future__ import absolute_import
from .pyaErrTemplate import PyaErrTemplate

def warn(w):
  """
    Parameters:
      - `w` - Something which can be converted into a string.
              This may especially be a PyA exception, which is treated as a warning here.
    
    Warnings are printed to stdout.
  """
  print("--------------------")
  print("| PyA User warning |")
  print("--------------------")
  if isinstance(w, PyaErrTemplate):
    print(w.__str__(head=False))
    return
  print(str(w))


# -*- coding: utf-8 -*-
import numpy
import PyAstronomy.funcFit as fuf
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.modelSuite.XTran import _ZList

try:
    from . import occultnl
    _importOccultnl = True
except ImportError:
    _importOccultnl = False


class MandelAgolNLLC(_ZList, fuf.OneDFit):
    r"""
    Calculate and fit analytical transit light-curves using the formulae
    provided by Mandel & Agol 2002. This model uses the non-linear
    limb-darkening prescription by Claret 2011:

    .. math :: \frac{I(\mu)}{I(1)}= 1 - \sum_{n=1}^{4}{a_n(1-\mu^{n/2})}

    .. note :: The computation of transit light curves
               is done using the external *occultnl* FORTRAN library.
               This library must be compiled manually using SciPy's f2py
               wrapper (http://www.scipy.org/F2py). Simply go to the
               *forTrans* directory of the source distribution of PyAstronomy,
               then invoke

               f2py -c occultnl.pyf occultnl.f



    *Model parameters*:
      - `p` - Radius ratio between planet and star.
      - `a` - Semi-major axis of planetary orbit [stellar radii].
      - `i` - Inclination of orbit in degrees (90 deg is *edge on* view).
      - `a1` - Non-Linear limb-darkening coefficient.
      - `a2` - Non-Linear limb-darkening coefficient.
      - `a3` - Non-Linear limb-darkening coefficient.
      - `a4` - Non-Linear limb-darkening coefficient.
      - `T0` - Time offset of transit center.
      - `per` - Period of planetary orbit.
      - `b` - Describes the flux ratio between a stellar companion and the main star (default is 0).

    """

    def __init__(self):
        _ZList.__init__(self, "circular")
        if not _importOccultnl:
            raise (PE.PyARequiredImport("Could not import required shared object library 'occultquad.so'",
                                        solution=["Invoke PyA's install script (setup.py) with the --with-ext option.",
                                                  "Go to 'forTrans' directory of PyAstronomy and invoke\n    f2py -c occultnl.pyf occultnl.f"]
                                        ))

        fuf.OneDFit.__init__(
            self, ["p", "a", "i", "a1", "a2", "a3", "a4", "T0", "per", "b"])
        self.freeze(["p", "a", "i", "a1", "a2", "a3", "a4", "T0", "per", "b"])
        self.setRootName("MandelAgolNL")

        self.__zlist = None

    def evaluate(self, time):
        r"""
        Calculate a light curve according to the analytical models
        given by Mandel & Agol 2002.

        Parameters
        ----------
        time : array
            An array of time points at which the light curve
            shall be calculated.

        .. note:: time = 0 -> Planet is exactly in the line of sight (phase = 0).

        Returns
        -------
        Model : array
            The analytical light curve is stored in the property `lightcurve`.
        """

        # Translate the given parameters into an orbit and, finally,
        # into a projected, normalized distance (z-parameter)
        self._calcZList(time - self["T0"])

        # Use occultquad Fortran library to compute flux decrease
        result = occultnl.occultnl(self["p"], self["a1"], self["a2"], self["a3"],
                                   self["a4"], self._zlist[self._intrans])

        df = numpy.zeros(len(time))
        df[self._intrans] = (1.0 - result[0])

        self.lightcurve = (1.-df)*1./(1.+self["b"]) + self["b"]/(1.0+self["b"])

        return self.lightcurve


MandelAgolNLLC_Rebin = fuf.turnIntoRebin(MandelAgolNLLC)



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyaC
import numpy as np
from PyAstronomy.pyaC import ImportCheck
import six.moves as smo


def generalizedESD(x, maxOLs, alpha=0.05, fullOutput=False, ubvar=False):
    """
    Carry out a Generalized ESD Test for Outliers.

    The Generalized Extreme Studentized Deviate (ESD) test for
    outliers can be used to search for outliers in a univariate
    data set, which approximately follows a normal distribution.
    A description of the algorithm is, e.g., given at
    `Nist <http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm>`_
    or [Rosner1983]_.

    Parameters
    ----------
    maxOLs : int
        Maximum number of outliers in the data set.
    alpha : float, optional
        Significance (default is 0.05).
    fullOutput : boolean, optional
        Determines whether additional return values
        are provided. Default is False.
    ubvar : boolean, optional
        If True, an unbiased estimate for the variance will be used in the
        calculation; this provides compatibility with the R implementation of
        NIST. If False (default), the maximum-likelihood variance estimate
        will be used.
    
    Returns
    -------
    Number of outliers : int
        The number of data points characterized as
        outliers by the test.
    Indices : list of ints
        The indices of the data points found to
        be outliers.
    R : list of floats, optional
        The values of the "R statistics". Only provided
        if `fullOutput` is set to True.
    L : list of floats, optional
        The lambda values needed to test whether a point
        should be regarded an outlier. Only provided
        if `fullOutput` is set to True.  

    """
    ImportCheck(["scipy"], required=["scipy"])
    from scipy.stats import t
    if maxOLs < 1:
        raise(PE.PyAValError("Maximum number of outliers, `maxOLs`, must be > 1.",
                             solution="Specify, e.g., maxOLs = 2"))
    import numpy.ma as ma
    xm = ma.array(x)
    n = len(xm)
    # Compute R-values
    R = []
    L = []
    minds = []
    for i in smo.range(maxOLs + 1):
        # Compute mean and std of x
        xmean = xm.mean()
        xstd = xm.std(ddof=int(ubvar))
        # Find maximum deviation
        rr = np.abs((xm - xmean) / xstd)
        minds.append(np.argmax(rr))
        R.append(rr[minds[-1]])
        if i >= 1:
            p = 1.0 - alpha / (2.0 * (n - i + 1))
            perPoint = t.ppf(p, n - i - 1)
            L.append((n - i) * perPoint /
                     np.sqrt((n - i - 1 + perPoint**2) * (n - i + 1)))
        # Mask that value and proceed
        xm[minds[-1]] = ma.masked
    # Remove the first entry from R, which is of
    # no meaning for the test
    R.pop(-1)
    # Find the number of outliers
    ofound = False
    for i in smo.range(maxOLs - 1, -1, -1):
        if R[i] > L[i]:
            ofound = True
            break
    # Prepare return value
    if ofound:
        if not fullOutput:
            # There are outliers
            return i + 1, minds[0:i + 1]
        else:
            return i + 1, minds[0:i + 1], R, L, minds
    else:
        # No outliers could be detected
        if not fullOutput:
            # There are outliers
            return 0, []
        else:
            return 0, [], R, L, minds


def pointDistGESD(x, maxOLs, alpha=0.05, ubvar=False):
    """
    Search for outliers by comparing distance of adjacent points.

    This method computes the "distance" between adjacent points, e.g.,
    d = x[i+1] - x[i]. It then uses :py:func:`PyAstronomy.pyasl.generalizedESD`
    to search for outliers in the list of distances. A point will be
    characterized as being an outlier, if (and only if) the distance
    to its left *and* right neighbor are abnormal.

    Parameters
    ----------
    x : array
        The data set to be searched for outliers.
    maxOLs : int
        The number of outliers. Note that the number
        passed to `generalizedESD` is actually 2*`maxOLs`.
    alpha : float, optional
        The significance level to be used in applying
        `generalizedESD`. The default is 0.05.
    ubvar : boolean, optional
        If True, an unbiased estimate for the variance will be used in the
        calculation; this provides compatibility with the R implementation of
        NIST. If False (default), the maximum-likelihood variance estimate
        will be used.

    Returns
    -------
    n : int
        The number of outliers detected.
    Indices : list of ints
        The indices of the points found to
        be outliers.
    """
    # Get the distances
    ds = x[1:] - x[0:-1]
    # Apply the generalized ESD to the distances
    r = generalizedESD(ds, maxOLs=maxOLs * 2, alpha=alpha, fullOutput=False, ubvar=ubvar)
    # Detect outliers (distance to left AND right should
    # be abnormal).
    oll = []
    for i in range(len(r[1])):
        # Check whether also the distance to the right
        # if the current point is abnormal
        if r[1][i] + 1 in r[1]:
            oll.append(r[1][i] + 1)
    return len(oll), oll


def polyResOutlier(x, y, deg=0, stdlim=3.0, controlPlot=False, fullOutput=False, mode="both"):
    """
    Simple outlier detection based on residuals.

    This algorithm fits a polynomial of the specified degree
    to the data, subtracts it to find the residuals, determines the
    standard deviations of the residuals, and, finally,
    identifies all points with residuals further than the
    specified number of standard deviations from the fit.

    Parameters
    ----------
    x, y : arrays
        The abscissa and ordinate of the data.
    deg : int, optional
        The degree of the polynomial to be fitted.
        The default is 0, i.e., a constant.
    stdlim : float, optional
        The number of standard deviations acceptable
        for points not categorized as outliers.
    mode : string, {both, above, below}
        If 'both' (default), outliers may be located on
        both sides of the polynomial. If 'above/below', outliers
        are only expected above/below it. 
    controlPlot : boolean, optional
        If True, a control plot will be generated
        showing the location of outliers (default is
        False).
    fullOutput : boolean, optional
        If True, the fitted polynomial and the resulting
        model will be returned.

    Returns
    -------
    indiin : array
        The indices of the points *not* being categorized
        as outliers.
    indiout : array
        Indices of the oulier points.
    p : array, optional
        Coefficients of the fitted polynomial (only returned if
        `fullOutput` is True).
    model : array, optional
        The polynomial model (only returned if
        `fullOutput` is True).
    """
    if len(x) < deg + 1:
        raise(PE.PyAValError("Only " + str(len(x)) + " points given to fit a polynomial of degree " + str(deg) + ".",
                             solution="Use more points and/or change degree of polynomial.",
                             where="polyResOutlier"))
    if len(x) != len(y):
        raise(PE.PyAValError("x and y need to have the same length.",
                             solution="Check the lengths of the input arrays.",
                             where="polyResOutlier"))
    if deg < 0:
        raise(PE.PyAValError("Polynomial degree must be > 0.",
                             where="polyResOutlier"))

    p = np.polyfit(x, y, deg)
    model = np.polyval(p, x)
    residuals = y - model

    std = np.std(residuals)
    # Find points too far off
    if mode == 'both':
        # Outliers above and/or below the curve
        indi = np.where(np.abs(residuals) >= stdlim * std)[0]
    elif mode == 'above':
        indi = np.where(residuals >= stdlim * std)[0]
    elif mode == 'below':
        indi = np.where(residuals <= -stdlim * std)[0]
    else:
        raise(PE.PyAValError("No such mode: " + str(mode),
                             where="polyResOutlier",
                             solution="Use any of 'both', 'above', or 'below'."))
    indiin = pyaC.invertIndexSelection(residuals, indi)
    if controlPlot:
        # Produce control plot
        import matplotlib.pylab as plt
        plt.title(
            "polyResOutlier control plot (red: potential outliers, green: polynomial model)")
        plt.plot(x, y, 'b.')
        s = np.argsort(x)
        plt.plot(x[s], model[s], 'g--')
        plt.plot(x[indi], y[indi], 'rp')
        plt.show()

    if fullOutput:
        return indiin, indi, p, model
    return indiin, indi


def slidingPolyResOutlier(x, y, points, count=1, deg=0, stdlim=3.0, controlPlot=False, dx=1, mode='both'):
    """
    Outlier detection based on polynomial fit in sliding box.

    This algorithm fits a polynomial of the specified degree
    to a sliding chunk of the data, subtracts it to find the
    residuals, determines the standard deviations of the residuals,
    and, finally, identifies all points with residuals further
    than the specified number of standard deviations from the fit.

    The length of the chunk is determined by `points`. In each step,
    the chunk is advanced by `dx` data points (default is one). To be finally
    maked as an outlier, a point must be detected as an outlier in at least
    `count` instances, when the chunk slides over it. By default, a single
    such detection is sufficient to establish its outlier status.  

    Parameters
    ----------
    x, y : arrays
        The abscissa and ordinate of the data.
    points : int
        Number of points for the sliding box
    count : int, optional
        Number of "slides" in which the point shall
        deviate from the fit by the stdlim
    deg : int, optional
        The degree of the polynomial to be fitted.
        The default is 0, i.e., a constant.
    stdlim : float, optional
        The number of standard deviations acceptable
        for points not categorized as outliers.
    mode : string, {both, above, below}
        If 'both' (default), outliers may be located on
        both sides of the polynomial. If 'above/below', outliers
        are only expected above/below it. 
    controlPlot : boolean, optional
        If True, a control plot will be generated
        showing the location of outliers (default is
        False).
    dx : int, optional
        The number of data points by which the chunk
        is advanced in each step.

    Returns
    -------
    indiin : array
        The indices of the points *not* categorized
        as outliers.
    indiout : array
        Indices of the oulier points.
    """
    if len(x) < deg + 1:
        raise(PE.PyAValError("Only " + str(len(x)) + " points given to fit a polynomial of degree " + str(deg) + ".",
                             solution="Use more points and/or change degree of polynomial.",
                             where="slidingPolyResOutlier"))
    if len(x) != len(y):
        raise(PE.PyAValError("x and y need to have the same length.",
                             solution="Check the lengths of the input arrays.",
                             where="slidingPolyResOutlier"))
    if deg < 0:
        raise(PE.PyAValError("Polynomial degree must be > 0.",
                             where="slidingPolyResOutlier"))

    good = np.ones(len(x)) * count

    if controlPlot:
        import matplotlib.pylab as plt
        # Produce control plot
        plt.close()
        plt.cla()
        plt.plot(x, y, 'b.-')

    for i in range(0, len(x) - points + 1, dx):
        # Indices of the chunk
        gi0 = np.arange(i, i + points)
        # Exclude points that have been already discarded
        gi = gi0[good[gi0] > 0]

        iin, iout = polyResOutlier(
            x[gi], y[gi], deg=deg, stdlim=stdlim, mode=mode)

        good[gi[iout]] -= 1

    indiout = np.where(good <= 0)
    indiin = np.where(good > 0)

    if controlPlot:
        plt.plot(x[indiout], y[indiout], 'ro')
        plt.show()

    return indiin[0], indiout[0]



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyasl
from .keplerOrbit import KeplerEllipse
from .astroTimeLegacy import helio_jd, daycnv
from . import observatory as pyaobs
from . import eq2hor
from . import sunpos
from . import twilight
from .moonpos import moonpos
from .moonphase import moonphase
from .angularDistance import getAngDist
from .cardinalPoint import getCardinalPoint
import sys
from . import airmass
from . import localtime
import six
import six.moves as smo
import scipy.optimize as sco


def isInTransit(time, T0, period, halfDuration, boolOutput=False, secin=False):
    r"""
    Check whether time is inclosed by transit interval.

    This function uses the given ephemerides (T0, period, and
    halfDuration) to check whether the time point(s)
    specified by `time` are within a transit window or not.
    The edges of the window are counted as transit times.

    Parameters
    ----------
    time : float or array like
        The time(s) which are to be checked.
    T0 : float
        The time reference point (center of transit).
    period : float
        The orbital period.
    halfDuration : float
        The half-duration of the event.
        Must have same units as `time`.
    boolOutput : boolean, optional
        If set True and `time` is an array, the function will
        return a bool array holding True for time points in-
        and False for time points out-of-transit.
    secin : boolean, optional
        If True, also points associated with the secondary
        transit (around phase 0.5) will be counted as falling
        into the transit window. Default is False.

    Returns
    -------
    inTransit : boolean or array of int
        If `time` was a float, the return value is a boolean,
        which is True if the give time falls into a transit
        interval and False otherwise.
        If `time` was given as an array, the return value is
        an array holding the indices of those time points,
        which fall into a transit window. The `boolOutput`
        option may be used to obtain a boolean array holding
        True for in-transit points.

    """
    if halfDuration > period / 2.0:
        raise (
            PE.PyAValError(
                "The half-duration is longer than half the period. This cannot be true.",
                where="isInTransit",
            )
        )
    if (period <= 0.0) or (halfDuration <= 0.0):
        raise (
            PE.PyAValError(
                "Both period and half-duration must be larger 0.", where="isInTransit"
            )
        )
    absPhase = np.array(np.abs((np.array(time) - T0) / period), ndmin=1)
    absPhase -= np.floor(absPhase)
    dPhase = halfDuration / period
    isIn = np.logical_or(absPhase <= dPhase, absPhase >= (1.0 - dPhase))
    if secin:
        isIn = isIn | (np.abs(absPhase - 0.5) < dPhase)
    indi = np.where(isIn)[0]
    if isinstance(time, float):
        return len(indi) == 1
    if boolOutput:
        return isIn
    return indi


def ingressDuration(sma, rp, rs, inc, period):
    r"""
    Calculate the ingress (and egress) duration of the transit.

    The routine calculates the transit in- and egress duration assuming
    a spherical star and planet and a circular planetary orbit.

    It evaluates the expression

    .. math::

        T_I = \frac{P}{2 \pi} \left( \arcsin\left(\frac{\sqrt{(R_s+R_p)^2 - b^2}}{a \sin(i)} \right) - \arcsin\left(\frac{\sqrt{(R_s-R_p)^2 - b^2}}{a \sin(i)} \right)  \right)

    is evaluated, where b is the impact parameter (:math:`b=a \cos(i)`).

    .. note:: The units of the ingress duration are the same as the units
              of the input orbital period.

    Parameters
    ----------
    sma : float
        The semi-major axis in AU.
    rp : float
        The planetary radius in Jovian radii.
    rs : float
        The stellar radius in solar radii.
    inc : float
        The orbital inclination in degrees.
    period : float
        The orbital period.

    Returns
    -------
    Ingress duration : float
        The duration of in- and egress (same units as `period`).
    """
    from PyAstronomy import constants as PC

    c = PC.PyAConstants()
    # Calculate the impact parameter
    impact = (sma * c.AU) * np.cos(inc / 180.0 * np.pi)
    imsin = (sma * c.AU) * np.sin(inc / 180.0 * np.pi)
    # Calculate the geometric transit duration
    idur = (period / (2.0 * np.pi)) * (
        np.arcsin(np.sqrt((rs * c.RSun + rp * c.RJ) ** 2 - impact**2) / imsin)
        - np.arcsin(np.sqrt((rs * c.RSun - rp * c.RJ)
                    ** 2 - impact**2) / imsin)
    )
    return idur


def ingressDuration_Rs(sma, rprs, inc, period):
    r"""
    Calculate transit duration

    Invokes :py:func:`ingressDuration` after parameter transformation.

    Parameters
    ----------
    sma : float
        Semi-major axis [stellar radii]
    rprs : float
        Planet-to-star radius ratio (Rp/Rs)
    inc : float
        Orbital inclination [deg]
    period : float
        Orbital period

    Returns
    -------
    Ingress duration : float
        The duration of in- and egress (same units as `period`).
    """
    from PyAstronomy import constants as PC

    c = PC.PyAConstants()

    rs = c.RSun
    return ingressDuration(sma * rs / c.AU, rprs * rs / c.RJ, 1.0, inc, period)


def transitDuration(sma, rp, rs, inc, period, exact=False):
    r"""
    Calculate the transit duration.

    The routine calculates the transit duration assuming
    a spherical star and planet and a circular planetary orbit.

    It evaluates the expression

    .. math::

        T_D = \frac{P}{\pi} \arcsin\left(\frac{\sqrt{(R_s+R_p)^2 - b^2}}{a} \right)

    where P is the orbital period, b the impact parameter (:math:`b=a \cos(i)`),
    and a the semi-major axis.

    If the `exact` flag is set True, the slightly more accurate expression

    .. math::

        T_D = \frac{P}{\pi} \arcsin\left(\frac{\sqrt{(R_s+R_p)^2 - b^2}}{a \sin(i)} \right)

    is evaluated, where the additional sine term accounts for orbit curvature.

    .. note:: The units of the transit duration are the same as the units
              of the input orbital period.

    Parameters
    ----------
    sma : float
        The semi-major axis in AU.
    rp : float
        The planetary radius in Jovian radii.
    rs : float
        The stellar radius in solar radii.
    inc : float
        The orbital inclination in degrees.
    period : float
        The orbital period.
    exact : boolean, optional
        If True, a slightly more accurate expression is evaluated.
        The default is False.

    Returns
    -------
    Transit duration : float
        The duration of the transit (same units as `period`).
    """
    from PyAstronomy import constants as PC

    c = PC.PyAConstants()
    # Calculate the impact parameter
    impact = (sma * c.AU) * np.cos(inc / 180.0 * np.pi)
    # Correction factor (inclination dependent factor)
    cfac = 1.0 if not exact else np.sin(inc / 180.0 * np.pi)
    # Calculate the geometric transit duration
    dur = (period / np.pi) * np.arcsin(
        np.sqrt((rs * c.RSun + rp * c.RJ) ** 2 -
                impact**2) / (sma * c.AU * cfac)
    )
    return dur


def transitDuration_Rs(sma, rprs, inc, period, exact=False):
    r"""
    Calculate transit duration

    Invokes :py:func:`transitDuration` after parameter transformation.

    Parameters
    ----------
    sma : float
        Semi-major axis [stellar radii]
    rprs : float
        Planet-to-star radius ratio (Rp/Rs)
    inc : float
        Orbital inclination [deg]
    period : float
        Orbital period
    exact : boolean, optional
        If True, a slightly more accurate expression is evaluated.
        The default is False.

    Returns
    -------
    Transit duration : float
        The duration of the transit (same units as `period`).
    """
    from PyAstronomy import constants as PC

    c = PC.PyAConstants()

    rs = c.RSun
    return transitDuration(sma * rs / c.AU, rprs * rs / c.RJ, 1.0, inc, period, exact)


def transit_T1_T4_ell(sma, rprs, inc, period, tau, e, w, transit="p"):
    r"""
    Calculate first to fourth contact times for elliptical orbit.

    The contact times are numerically estimated using the algorithms
    provided by :py:class:`KeplerEllipse`.

    Parameters
    ----------
    sma : float
        Semi-major axis [stellar radii]
    rprs : float
        Planet-to-star radius ratio (Rp/Rs)
    inc : float
        Orbital inclination [deg]
    period : float
        Orbital period
    tau : float
        Time of periastron passage
    e : float
        Eccentricity
    w : float
        Argument of periastron
    transit : string, {p, s}, optional
        Determines whether contact point times for primary (p, default)
        or secondary (s) transit are to be calculated.

    Returns
    -------
    T1-T4 : tuple of floats
        Contact times (arbitrary epoch). Note that contact time may not exist
        in the case of grazing transits or no transit at all. None is returned
        for undefined contact times.
    """

    if not transit in ["p", "s"]:
        raise (
            PE.PyAValError(
                "transit must be 'p' or 's' ('" +
                str(transit) + "' was given).",
                where="transit_T1_T4_ell",
            )
        )

    ke = KeplerEllipse(sma, period, e=e, tau=tau, w=w, i=inc)

    def r(t):
        # R in terms of Rs
        p = ke.xyzPos(t)
        r = np.sqrt(p[0] ** 2 + p[1] ** 2)
        return r

    # Estimate transit time (central transit time in case of circular orbit)
    t0sec, t0prim = ke.yzCrossingTime(ordering="z")
    if transit == "p":
        t0 = t0prim
    else:
        t0 = t0sec
    t0 = t0 % period

    # Get nodes positions and times
    na, nd = ke.xyzNodes_LOSZ(getTimes=True)
    # Times between which T1 and T4 must occur
    if transit == "p":
        lims = (nd[1], na[1])
    else:
        lims = (na[1], nd[1])
    lims = [l % period for l in lims]
    if lims[0] > lims[1]:
        lims[1] += period
    if t0 < lims[0]:
        t0 += period

    try:
        t1 = sco.brentq(lambda t: r(t) - (1 + rprs), lims[0], t0)
    except ValueError:
        t1 = None
    try:
        t2 = sco.brentq(lambda t: r(t) - (1 - rprs), lims[0], t0)
    except ValueError:
        t2 = None
    try:
        t3 = sco.brentq(lambda t: r(t) - (1 - rprs), t0, lims[1])
    except ValueError:
        t3 = None
    try:
        t4 = sco.brentq(lambda t: r(t) - (1 + rprs), t0, lims[1])
    except ValueError:
        t4 = None
    return t1, t2, t3, t4


def collectTransitDataFromDB(pn, dbp=("nexa", "eeu"), verbose=False):
    r"""
    Get transit timing data from database

    Parameters
    ----------
    pn : string
        Planet name (e.g., 'WASP-12 b')
    dbp : tuple of strings
        A tuple with the strings 'nexa' and 'eeu', which specify the priorities
        of the exoplanet databases. Data is collected from the first DB, which returns
        useful results.
    verbose : boolean, optional
        If True, information about the progress is printed. Default is False.

    Returns
    -------
    Transit data : dictionary
        A dictionary with keys: ra, dec, T0, orbPer, orbInc, SMA, RpJ, RsSun, and optionally Tdur.
        Suitable to be used with the `transitTimes` function.
    DB name : string, {nexa, eeu}
        The name of the database from which the data were collected.
    """

    def colfromnexa(pn):
        nexa = pyasl.NasaExoplanetArchive()
        d = nexa.selectByPlanetName(pn)
        m = {
            "ra": "ra",
            "dec": "dec",
            "T0": "pl_tranmid",
            "orbPer": "pl_orbper",
            "orbInc": "pl_orbincl",
            "SMA": "pl_orbsmax",
            "RpJ": "pl_radj",
            "RsSun": "st_rad",
            "Tdur": "pl_trandur",
        }
        r = {k: d[m[k]] for k in list(m)}
        # Convert duration into days
        r["Tdur"] /= 24.0
        return r

    def colfromeu(pn):
        eu = pyasl.ExoplanetEU2()
        d = eu.selectByPlanetName(pn)
        m = {
            "ra": "ra",
            "dec": "dec",
            "T0": "tzero_tr",
            "orbPer": "orbital_period",
            "orbInc": "inclination",
            "SMA": "semi_major_axis",
            "RpJ": "radius",
            "RsSun": "star_radius",
        }
        r = {k: d[m[k]] for k in list(m)}
        r["Tdur"] = None
        return r

    dbm = {"nexa": colfromnexa, "eeu": colfromeu}
    for db in dbp:
        try:
            r = dbm[db](pn)
            try:
                r["ra"] = float(r["ra"])
            except ValueError:
                # Convert into deg
                r["ra"], r["dec"] = pyasl.coordsSexaToDeg(
                    r["ra"] + " " + r["dec"])
            m = ["ra", "dec", "T0", "orbPer", "orbInc", "SMA", "RpJ", "RsSun"]
            for k in m:
                if not np.isfinite(r[k]):
                    raise (PE.PyAValError(f"Value {k} not defined (finite)"))
        except Exception as e:
            if verbose:
                print(f"Failed to collect planet data from DB {db}")
                print(e)
        else:
            return r, db

    return None, None


def transitTimes(
    tmin,
    tmax,
    planetData,
    obsOffset=0.0,
    hjd=True,
    observatory=None,
    lon=None,
    lat=None,
    alt=None,
    minAltitude=None,
    showTwilight="all",
    moonDist=None,
    nexaInput=False,
    fileOutput=None,
    psT0=0.0,
):
    r"""
    Calculate transit times for a given planet and a given period of time.

    The `planetData` dictionary must contain the following information:

    =======  =================================
    Key      Value
    -------  ---------------------------------
    ra       Right ascension of object [deg]
    dec      Declination of object [deg]
    T0       Time reference point (HJD)
    orbPer   Orbital period [d]
    orbInc   Orbital inclination [deg]
    SMA      Semi-major axis [AU]
    RpJ      Planetary radius [Jovian radii]
    RsSun    Stellar Radius [solar]
    Tdur     OPTIONAL, Transit duration [d]
    =======  =================================

    If the transit duration (Tdur) is not given, the duration will
    be estimated using pyasl's `transitDuration` function.

    .. note:: The input times (`tmin` and `tmax`) are expected in JD (UT).
              Input time will be calculated to HJD.
              Time output is in HJD.

    Parameters
    ----------
    tmin : float
        Start of time interval in Julian days (UT).
    tmax : float
        End of time interval in Julian days (UT).
    planetData: dictionary or string
        Either a dictionary containing the parameters of the exoplanet
        for which the transit times should be calculated or the name of the
        planet (e.g., 'WASP-12 b'). In the case of a string, an attempt will be
        made to download the data from a database.
        The required keys for the dictionary are specified above.
    obsOffset : float, optional
        Specifies additional time before AND after the transit in DAYS.
        This is useful if the observation should start and end
        some time before and after the actual transit.
    hjd : boolean, optional
        If True (default), the given Julian dates specifying the time
        interval (`tmin` and `tmax`) are automatically
        converted into the heliocentric frame (HJD).
    observatory : string, optional
        If given, pyasl's `observatory` function will be used to automatically
        resolve the name and obtain longitude, latitude, and altitude
        of the observatory.
        If `observatory` is given, `lon`, `lat`, and `alt` must not be specified.
    lon : float, optional
        East longitude of the observatory given in DEGREES.

        Longitude is positive in EASTWARD direction.
        If LON is not given, transitTimes will only return beginning and end
        of the observation and the time of mid transit.
    lat : float, optional
        Latitude of the observatory given in DEGREES
        (positive in NORTHWARD direction).
    alt : float, optional
        Altitude of the observatory given in METER.
    minAltitude : float, optional
        Minimum altitude of the object in DEGREES.

        If a minimum altitude is given, only transits for which the
        object is above the given altitude during the ENTIRE
        observation are included in the list
        created by `transitTimes`. Note that `minAltitude` can
        only be used if the observer's location has been specified
        either via `observatory` or via `lon`, `lat`, and `alt`.
    showTwilight : string, optional, {"all", "civil", "nautical", "astronomical", "night"}
        Specifies the twilight acceptable during the observation.
        By default all twilight conditions are acceptable.

        Only the transits for which the ENTIRE observation
        occurs during the specified or darker twilight conditions
        are listed.

        The choices are:
          - "all": all transits are shown (even during day)
          - "civil": only transits during civil twilight and better are shown
          - "nautical": only transits during nautical twilight and better are shown
          - "astronomical": only transits during astronomical twilight and better are shown
          - "night": only transits during night are shown

        Note that this can only have an effect, if the observer's location is
        specified.
    moonDist : float
        Minimum distance between the Moon and the target in DEGREES.
        By default all Moon distances are acceptable (moonDist=0.0).

        Only observations are listed for which the angular distance between
        the Moon and the target is larger
        than `moonDist` during the ENTIRE observation.

        Note that this can only have an effect, if the observer's location is
        specified.
    psT0 : float, optional
        Phase-shift the transit midpoint. This may be useful to determine, e.g.,
        eclipse times. In this case, psT0=0.5 may be suitable, which shifts T0
        by 0.5 orbital periods. Default is 0.0
    fileOutput : string or file, optional
        If a string is given, a file with the name will be created
        and the output will be written to that file. If a (writable)
        file object is given, the output will be written to that
        file. In both cases, no output will be given on screen.

    Returns
    -------
    Transit times : dictionary
        Returns a dictionary containing the transit details. The dictionary key
        is a running number (starting with one), which is equivalent to that
        listed in the first column of the table.

        For each transit, the function returns a dictionary with the transit
        details.

        If the observer's location was not specified, the dictionary has the
        following keys:

          ============    ====================================================
          Key             Description
          ------------    ----------------------------------------------------
          Planet name     Name of the planet
          Tmid            HJD of transit center
          Transit jd      Array giving JD of start, mid-time, and end of
                          transit.
          Obs jd          Array specifying the HJD of the start, center and
                          end of the observation.
          Obs cal         Equivalent to 'Obs jd', but in the form of the
                          calendar date. In particular, for each date, a list
                          containing [Year, month, day, fractional hours]
                          is given.

                          **Below follows optional output only present**
                          **if the observer's location is known**

          Obs coord       East longitude [deg], latitude [deg], and
                          altitude [m] of the observatory.
          Sun ra          Right ascension of the Sun at center of
                          observation.
          Sun dec         Declination of the Sun at center of
                          observation.
          Sun alt         Altitude of the Sun [deg] at begin, center, and
                          end of the observation.
          Sun az          Azimuth if the Sun [deg] at begin, center, and
                          end of the observation.
          Moon phase      Array giving lunar phase (in percent) at start,
                          center, and end of the observation.
          Moon AD         Angular distance between the target and the Moon
                          at begin, center, and end of the observation [deg].
          Moon ra         Right ascension of the Moon at begin, center, and
                          end of the observation [deg].
          Moon dec        Declination of the Moon at begin, center, and
                          end of the observation [deg].
          Star ra         Right ascension of the star [deg].
          Star dec        Declination of the star [deg].
          Star CP         Cardinal point of the star at begin, center, and
                          end of the observation.
          Star alt        Altitude of the star [deg] at begin, center, and
                          end of the observation.
          Star az         Azimuth of the star [deg] at begin, center, and
                          end of the observation.
          Twilight        The worst, i.e., brightest type of twilight
                          encountered during the observation.
          ============    ====================================================

    """

    if fileOutput is not None:
        oldStdout = sys.stdout
        if isinstance(fileOutput, six.string_types):
            sys.stdout = open(fileOutput, "w")
        else:
            sys.stdout = fileOutput

    try:
        if tmin >= tmax:
            raise (
                PE.PyAValError(
                    "The given time range is inconsistent (tmin >= tmax)",
                    where="transitTimes",
                    solution="Adapt tmin and tmax.",
                )
            )

        if isinstance(planetData, dict):
            # Copy input dictionary, because it may be changed
            planetData = planetData.copy()

            if nexaInput:
                pdin = planetData.copy()
                planetData = {}
                planetData["ra"] = pdin["ra"]
                planetData["dec"] = pdin["dec"]
                planetData["orbPer"] = pdin["pl_orbper"]
                planetData["T0"] = pdin["pl_tranmid"]
                planetData["orbInc"] = pdin["pl_orbincl"]
                planetData["SMA"] = pdin["pl_orbsmax"]
                planetData["RpJ"] = pdin["pl_radj"]
                planetData["RsSun"] = pdin["st_rad"]
                planetData["Tdur"] = pdin["pl_trandur"] / \
                    24  # Convert into days
                planetData["plName"] = pdin["pl_name"]
                if np.isnan(planetData["Tdur"]):
                    del planetData["Tdur"]
        elif isinstance(planetData, str):
            # Try to collect data from DBs
            pn = planetData
            print(f"Collecting transit parameters for {pn} from database")
            planetData, dbname = collectTransitDataFromDB(pn, verbose=False)
            if planetData is None:
                raise (
                    PE.PyAValError(
                        f"Could not download transit data for target '{pn}' from data base.",
                        where="transitTimes",
                        solution=["Check name of planet",
                                  "Specify parameter manually"],
                    )
                )
            planetData["plName"] = pn
            if (planetData["Tdur"] is None) or (not np.isfinite(planetData["Tdur"])):
                del planetData["Tdur"]

        # Check whether required keys are present
        reke = ["ra", "dec", "orbPer", "T0",
                "orbInc", "SMA", "RpJ", "RsSun", "plName"]
        msg = ""
        fail = False
        for key in reke:
            if not key in planetData:
                msg += "The required key '" + key + "' is missing in the input data!\n"
                fail = True
                continue
            if isinstance(planetData[key], (tuple(six.integer_types) + (float,))):
                if np.isnan(planetData[key]):
                    msg += (
                        "The required key '"
                        + key
                        + "' has NaN value in the input data!\n"
                    )
                    fail = True
        if fail:
            raise (
                PE.PyAValError(
                    "The input `planetData` is inappropriate:\n" + msg,
                    where="transitTimes",
                    solution="Specify all required input values.",
                )
            )

        if psT0 != 0.0:
            print()
            pst = planetData["orbPer"] * psT0
            print(f"Shifting T0 by {psT0} orbital periods, i.e., by {pst}")
            print(f"    T0 before shift = {planetData['T0']}")
            planetData["T0"] += pst
            print(f"    T0 after shift = {planetData['T0']}")
            print()

        print("-"*30)
        print("Adopted transit parameters")
        for k, v in planetData.items():
            print(f"    {k} = {v}")
        print("-"*30)

        # Object position [degrees]
        ra = planetData["ra"]
        dec = planetData["dec"]

        if hjd:
            # Convert input times into heliocentric frame
            # Using 'reduced' JD in calculation
            tmin = helio_jd(tmin - 2.4e6, ra, dec) + 2.4e6
            tmax = helio_jd(tmax - 2.4e6, ra, dec) + 2.4e6

        print("Specified time span")
        print(
            "Start date (DDDD-MM-YY and fractional hours): {0:4d}-{1:02d}-{2:02d} {3:6.3f}".format(
                *daycnv(tmin)
            )
        )
        print(
            "End date (DDDD-MM-YY and fractional hours): {0:4d}-{1:02d}-{2:02d} {3:6.3f}".format(
                *daycnv(tmax)
            )
        )
        print()

        # Transit parameters
        # Orbital period in days
        period = planetData["orbPer"]
        # Transit reference time (should be HJD)
        T0 = planetData["T0"]

        if not "Tdur" in planetData:
            # No duration specified in the data
            inc = planetData["orbInc"]  # deg
            sma = planetData["SMA"]  # au
            rp = planetData["RpJ"]  # Mjup
            rs = planetData["RsSun"]  # Msun
            dur = transitDuration(sma, rp, rs, inc, period)
            print(
                "Estimating transit duration using orbital inclination, semi-major axis,"
            )
            print("  planetary radius, and stellar radius")
        else:
            dur = planetData["Tdur"]

        print("Transit duration: ", dur * 24.0 * 60.0, " minutes")
        print(
            "Off-transit time before and after transit: ",
            obsOffset * 24.0 * 60.0,
            " minutes",
        )

        # First and last epoch contained in specified range
        trnum_start = np.floor((tmin - T0) / period)
        trnum_end = np.ceil((tmax - T0) / period)
        # Relevant transit epochs
        tr = np.arange(trnum_start, trnum_end, 1)

        if (observatory is not None) and (
            (lon is not None) or (lat is not None) or (alt is not None)
        ):
            raise (
                PE.PyAParameterConflict(
                    "You must either specify `observatory` OR `lon`, `lat`, and `alt`.",
                    where="transitTimes",
                    solution="Adapt function call.",
                )
            )

        if observatory is not None:
            # Resolve observatory string
            observatory_data = pyaobs.observatory(observatory)
            lon = observatory_data["longitude"]
            lat = observatory_data["latitude"]
            alt = observatory_data["altitude"]

        # Check if the observatory data are complete
        obsCompl = (lon is None) + (lat is None) + (alt is None)
        if (obsCompl == 1) or (obsCompl == 2):
            raise (
                PE.PyAValError(
                    "Observatory data is incomplete. `lon`, `lat`, and `alt` must all be specified.\n"
                    + "Current values are: lon = "
                    + str(lon)
                    + ", lat = "
                    + str(lat)
                    + ", alt = "
                    + str(alt),
                    where="transitTimes",
                    solution="Provide complete observatory information.",
                )
            )

        if (minAltitude is not None) and (lon is None):
            # Observer's location not given so minAltitude cannot have any effect
            raise (
                PE.PyAParameterConflict(
                    "The observer's location is not specified, but `minAltitude` is given.\n"
                    + "This parameter can only be used, if the observer's location is known.",
                    where="transitTimes",
                    solution="Either specify the observer's location or set `minAltitude` to None.",
                )
            )

        if (showTwilight != "all") and (lon is None):
            # Observer's location not given so showTwilight cannot have any effect
            raise (
                PE.PyAParameterConflict(
                    "The observer's location is not specified, but `showTwilight` is given.\n"
                    + "This parameter can only be used, if the observer's location is known.",
                    where="transitTimes",
                    solution='Either specify the observer\'s location or set `showTwilight` to "all".',
                )
            )

        if (
            (showTwilight != "all")
            and (showTwilight != "civil")
            and (showTwilight != "nautical")
            and (showTwilight != "astronomical")
            and (showTwilight != "night")
        ):
            # None of the possible choices for showTwilight have been used.
            raise (
                PE.PyAValError(
                    "Wrong keyword given for showTwilight.\n"
                    + "Current keyword is "
                    + showTwilight,
                    where="transitTimes",
                    solution="Select a valid keyword for showTwilight: `all', `civil', `nautical', `astronomical', or `night'.",
                )
            )

        if moonDist is None:
            # No limit on Moon distance
            moonDist = 0.0
        if (moonDist != 0.0) and (lon is None):
            # Observer's location not given so showTwilight cannot have any effect
            raise (
                PE.PyAParameterConflict(
                    "The observer's location is not specified, but `moonDist` is given.\n"
                    + "This parameter can only be used, if the observer's location is known.",
                    where="transitTimes",
                    solution="Either specify the observer's location or set `moonDist` to 0.0 or None.",
                )
            )

        if moonDist < 0.0:
            # Moon distance below zero does not make sense
            PE.warn(
                "The specified `moonDist' is below zero ("
                + str(moonDist)
                + ") which does not make sense.\n"
                + "It was changed to 0.0.\n"
                + "Please use a value >= 0.0 or None if specifying `moonDist'."
            )

        print()
        if np.logical_and(lon != None, lat != None):
            print(
                "No. Tmid [HJD]      Obs. start [UT] [ALT, DIR(AZI)]     Transit mid [UT] [ALT, DIR(AZI)]     Obs. end [UT] [ALT, DIR(AZI)]   twilight"
                + " (SUN ALT)                   moon distance     moon phase"
            )
        else:
            print(
                "No. Tmid [HJD]      Obs. start [UT]    Transit mid [UT]   Obs. end [UT]"
            )

        allData = {}
        trcounter = 1
        for i in tr:
            trData = {}
            # Get times
            Tmid = T0 + float(i) * period
            if (Tmid < tmin) or (Tmid > tmax):
                # This may happen because the transit may occur in the first
                # relevant epoch but still before tmin. Likewise for tmax.
                continue
            obs_start_hjd = Tmid - (dur / 2.0) - obsOffset
            obs_start = daycnv(obs_start_hjd)
            obs_mid = daycnv(Tmid)
            obs_end_hjd = Tmid + (dur / 2.0) + obsOffset
            obs_end = daycnv(obs_end_hjd)
            time_temp = np.array([obs_start_hjd, Tmid, obs_end_hjd])
            transit_only = np.array(
                [Tmid - (dur / 2.0), Tmid, Tmid + (dur / 2.0)])

            # Get visibility
            if (lon is not None) and (lat is not None):
                # Get alt/az of object for current transit
                altaz = eq2hor.eq2hor(
                    time_temp,
                    np.ones(time_temp.size) * ra,
                    np.ones(time_temp.size) * dec,
                    lon=lon,
                    lat=lat,
                    alt=alt,
                )
                # If minimum altitude is not fulfilled during observation,
                # do not show transit
                if minAltitude is not None:
                    minalt = np.where(altaz[0] >= minAltitude)[0]
                    if len(minalt) < time_temp.size:
                        # Skip this transit
                        continue
                # Get Sun position for current transit
                sunpos_radec = sunpos.sunpos(time_temp[1])
                sunpos_altaz = eq2hor.eq2hor(
                    time_temp,
                    np.ones(time_temp.size) * sunpos_radec[1],
                    np.ones(time_temp.size) * sunpos_radec[2],
                    lon=lon,
                    lat=lat,
                    alt=alt,
                )
                twi = twilight.twilightName(max(sunpos_altaz[0]))
                # Check type of twilight -> if requirement not fulfilled, don't show transit
                if showTwilight == "civil":
                    # Show civil or better
                    if twi == "day":
                        continue
                if showTwilight == "nautical":
                    # Show nautical or better
                    if twi == "day":
                        continue
                    if twi == "civil twilight":
                        continue
                if showTwilight == "astronomical":
                    # Show astronomical or better
                    if twi == "day":
                        continue
                    if twi == "civil twilight":
                        continue
                    if twi == "nautical twilight":
                        continue
                if showTwilight == "night":
                    # Only show night
                    if twi != "night":
                        continue

                # Get Moon position for current transit
                mpos = moonpos(time_temp)
                mdists = []
                for i in range(time_temp.size):
                    mdists.append(getAngDist(mpos[0][i], mpos[1][i], ra, dec))
                mdist = min(mdists)
                # Check Moon distance, if not fulfilled, neglect the transit
                if mdist < moonDist:
                    continue
                # Get lunar phase in percent
                moonpha = moonphase(time_temp) * 100.0
                print(
                    "%3d %10.5f   %2d.%2d. %2d:%02d    [%3d°,%s(%3d°)]      %2d.%2d. %2d:%02d     [%3d°,%s(%3d°)]      %2d.%2d. %2d:%02d  [%3d°,%s(%3d°)]   %18s (%3d°,%3d°,%3d°)   (%3d°,%3d°,%3d°)  %3d%%"
                    % (
                        trcounter,
                        Tmid,
                        obs_start[2],
                        obs_start[1],
                        np.floor(obs_start[3]),
                        (obs_start[3] - np.floor(obs_start[3])) * 60.0,
                        altaz[0][0],
                        getCardinalPoint(altaz[1][0]),
                        altaz[1][0],
                        obs_mid[2],
                        obs_mid[1],
                        np.floor(obs_mid[3]),
                        (obs_mid[3] - np.floor(obs_mid[3])) * 60.0,
                        altaz[0][1],
                        getCardinalPoint(altaz[1][1]),
                        altaz[1][1],
                        obs_end[2],
                        obs_end[1],
                        np.floor(obs_end[3]),
                        (obs_end[3] - np.floor(obs_end[3])) * 60.0,
                        altaz[0][2],
                        getCardinalPoint(altaz[1][2]),
                        altaz[1][2],
                        twi,
                        sunpos_altaz[0][0],
                        sunpos_altaz[0][1],
                        sunpos_altaz[0][2],
                        mdists[0],
                        mdists[1],
                        mdists[2],
                        np.max(moonpha),
                    )
                )
                # Save transit data
                trData["Tmid"] = Tmid
                trData["Obs jd"] = time_temp
                trData["Obs cal"] = [obs_start, obs_mid, obs_end]
                trData["Star ra"] = ra
                trData["Star dec"] = dec
                trData["Star alt"] = altaz[0]
                trData["Star az"] = altaz[1]
                trData["Sun ra"] = sunpos_radec[1]
                trData["Sun dec"] = sunpos_radec[2]
                trData["Sun alt"] = sunpos_altaz[0]
                trData["Sun az"] = sunpos_altaz[1]
                trData["Twilight"] = twi
                trData["Moon ra"] = mpos[0]
                trData["Moon dec"] = mpos[1]
                trData["Moon AD"] = mdist
                trData["Moon phase"] = moonpha
                trData["Star CP"] = [
                    getCardinalPoint(altaz[1][0]),
                    getCardinalPoint(altaz[1][1]),
                    getCardinalPoint(altaz[1][2]),
                ]

                trData["Obs coord"] = [lon, lat, alt]
            else:
                # If you do not specify the observer's location, return all transits of the object
                print(
                    "%3d %10.5f   %2d.%2d. %2d:%02d       %2d.%2d. %2d:%02d       %2d.%2d. %2d:%02d"
                    % (
                        trcounter,
                        Tmid,
                        obs_start[2],
                        obs_start[1],
                        np.floor(obs_start[3]),
                        (obs_start[3] - np.floor(obs_start[3])) * 60.0,
                        obs_mid[2],
                        obs_mid[1],
                        np.floor(obs_mid[3]),
                        (obs_mid[3] - np.floor(obs_mid[3])) * 60.0,
                        obs_end[2],
                        obs_end[1],
                        np.floor(obs_end[3]),
                        (obs_end[3] - np.floor(obs_end[3])) * 60.0,
                    )
                )
                trData["Tmid"] = Tmid
                trData["Obs jd"] = time_temp
                trData["Obs cal"] = [obs_start, obs_mid, obs_end]

            trData["Transit jd"] = transit_only
            trData["Planet name"] = planetData["plName"]
            allData[trcounter] = trData
            trcounter += 1

        if len(allData) == 0:
            print()
            print("------------------------------------------------------")
            print("!!! No transits found for the given restrictions. !!!")
            print("------------------------------------------------------")
            print()

    except:
        raise
    finally:
        if fileOutput is not None:
            if isinstance(fileOutput, six.string_types):
                sys.stdout.close()
            sys.stdout = oldStdout

    return allData


def transitVisibilityPlot(
    allData, markTransit=False, plotLegend=True, showMoonDist=True, print2file=False
):
    r"""
    Plot the visibility of transits.

    This function can conveniently be used with the output of
    the transitTimes function.

    Parameters
    ----------
    allData : dictionary
        Essentially the output of `transitTimes`.
        A dictionary mapping consecutive numbers (one per transit) to
        another dictionary providing the following keys:

          ============    ====================================================
          Key             Description
          ------------    ----------------------------------------------------
          Planet name     Name of the planet
          Transit jd      (Only if `markTransit is True)
                          Array giving JD of start, mid-time, and end of
                          transit.
          Obs jd          Array specifying the HJD of the start, center and
                          end of the observation.
          Obs cal         Equivalent to 'Obs jd', but in the form of the
                          calendar date. In particular, for each date, a list
                          containing [Year, month, day, fractional hours]
                          is given.
          Obs coord       East longitude [deg], latitude [deg], and
                          altitude [m] of the observatory.
          Star ra         Right ascension of the star [deg].
          Star dec        Declination of the star [deg].
          ============    ====================================================

        .. note:: To use the list created by transitTimes, the LONGITUDE and LATITUDE
                  of the observatory location must have been specified.
    markTransit : boolean, optional
        If True (default is False), the in-transit times will
        be clearly indicated in the plot.
        Note that this would not be the case otherwise, which is particularly
        important if extra off-transit time before and after the transit has been
        requested.
    showMoonDist : boolean, optional
        If True (default), the Moon distance will be shown.
    print2file : boolean or string, optional
        If True, the plot will be dumped to a png-file named:
        "transitVis-"[planetName].png. The default is False. If a string is given,
        it specifies the name of the output file.
    """
    from PyAstronomy.pyasl import _ic

    if not _ic.check["matplotlib"]:
        raise (
            PE.PyARequiredImport(
                "matplotlib is not installed.",
                where="transitVisibilityPlot",
                solution="Install matplotlib (http://matplotlib.org/)",
            )
        )

    import matplotlib
    import matplotlib.pylab as plt
    from mpl_toolkits.axes_grid1 import host_subplot
    from matplotlib.ticker import MultipleLocator
    from matplotlib.font_manager import FontProperties
    from matplotlib import rcParams

    rcParams["xtick.major.pad"] = 12

    if len(allData) == 0:
        raise (
            PE.PyAValError(
                "Input dictionary is empty",
                where="transitVisibilityPlot",
                solution=[
                    "Use `transitTimes` to generate input dictionary",
                    "Did you forget to supply observer's location?",
                    "If you used `transitTime`, you might need to change the call argument (e.g., times)",
                ],
            )
        )

    # Check whether all relevant data have been specified
    reqK = ["Obs jd", "Obs coord", "Star ra",
            "Star dec", "Obs cal", "Planet name"]
    if markTransit:
        reqK.append("Transit jd")
    missingK = []
    for k in reqK:
        if not k in allData[1]:
            missingK.append(k)
    if len(missingK) > 0:
        raise (
            PE.PyAValError(
                "The following keys are missing in the input dictionary: "
                + ", ".join(missingK),
                where="transitVisibilityPlot",
                solution="Did you specify observer's location in `transitTimes`?",
            )
        )

    fig = plt.figure(figsize=(15, 10))
    fig.subplots_adjust(left=0.07, right=0.8, bottom=0.15, top=0.88)
    ax = host_subplot(111)

    font0 = FontProperties()
    font1 = font0.copy()
    font0.set_family("sans-serif")
    font0.set_weight("light")
    font1.set_family("sans-serif")
    font1.set_weight("medium")

    for n in six.iterkeys(allData):
        # JD array
        jdbinsize = 1.0 / 24.0 / 10.0
        jds = np.arange(allData[n]["Obs jd"][0],
                        allData[n]["Obs jd"][2], jdbinsize)
        # Get JD floating point
        jdsub = jds - np.floor(jds[0])
        # Get alt/az of object
        altaz = eq2hor.eq2hor(
            jds,
            np.ones(jds.size) * allData[n]["Star ra"],
            np.ones(jds.size) * allData[n]["Star dec"],
            lon=allData[n]["Obs coord"][0],
            lat=allData[n]["Obs coord"][1],
            alt=allData[n]["Obs coord"][2],
        )
        # Get alt/az of Sun
        sunpos_altaz = eq2hor.eq2hor(
            jds,
            np.ones(jds.size) * allData[n]["Sun ra"],
            np.ones(jds.size) * allData[n]["Sun dec"],
            lon=allData[n]["Obs coord"][0],
            lat=allData[n]["Obs coord"][1],
            alt=allData[n]["Obs coord"][2],
        )

        # Define plot label
        plabel = "[%02d]  %02d.%02d.%4d" % (
            n,
            allData[n]["Obs cal"][0][2],
            allData[n]["Obs cal"][0][1],
            allData[n]["Obs cal"][0][0],
        )

        # Find periods of: day, twilight, and night
        day = np.where(sunpos_altaz[0] >= 0.0)[0]
        twi = np.where(np.logical_and(sunpos_altaz[0] > -18.0, sunpos_altaz[0] < 0.0))[
            0
        ]
        night = np.where(sunpos_altaz[0] <= -18.0)[0]

        if (len(day) == 0) and (len(twi) == 0) and (len(night) == 0):
            print()
            print(
                "transitVisibilityPlot - no points to draw for date %2d.%2d.%4d"
                % (
                    allData[n]["Obs cal"][0][2],
                    allData[n]["Obs cal"][0][1],
                    allData[n]["Obs cal"][0][0],
                )
            )
            print("Skip transit and continue with next")
            print()
            continue

        mpos = moonpos(jds)
        mpha = moonphase(jds)
        mpos_altaz = eq2hor.eq2hor(
            jds,
            mpos[0],
            mpos[1],
            lon=allData[n]["Obs coord"][0],
            lat=allData[n]["Obs coord"][1],
            alt=allData[n]["Obs coord"][2],
        )
        moonind = np.where(mpos_altaz[0] > 0.0)[0]

        if showMoonDist:
            mdist = getAngDist(
                mpos[0],
                mpos[1],
                np.ones(jds.size) * allData[n]["Star ra"],
                np.ones(jds.size) * allData[n]["Star dec"],
            )
            bindist = int((2.0 / 24.0) / jdbinsize)
            firstbin = np.random.randint(0, bindist)
            for mp in range(0, int(len(jds) / bindist)):
                bind = int(firstbin + float(mp) * bindist)
                ax.text(
                    jdsub[bind],
                    altaz[0][bind] - 1.0,
                    str(int(mdist[bind])) + r"$^\circ$",
                    ha="center",
                    va="top",
                    fontsize=8,
                    stretch="ultra-condensed",
                    fontproperties=font0,
                    alpha=1.0,
                )

        if markTransit:
            # Mark points within transit. These may differ from that pertaining to the
            # observation if an extra offset was given to provide off-transit time.
            transit_only_ind = np.where(
                np.logical_and(
                    jds >= allData[n]["Transit jd"][0],
                    jds <= allData[n]["Transit jd"][2],
                )
            )[0]
            ax.plot(
                jdsub[transit_only_ind],
                altaz[0][transit_only_ind],
                "g",
                linewidth=6,
                alpha=0.3,
            )

        if len(twi) > 1:
            # There are points in twilight
            linebreak = np.where((jdsub[twi][1:] - jdsub[twi][:-1]) > 2.0 * jdbinsize)[
                0
            ]
            if len(linebreak) > 0:
                plotrjd = np.insert(jdsub[twi], linebreak + 1, np.nan)
                plotdat = np.insert(altaz[0][twi], linebreak + 1, np.nan)
                ax.plot(plotrjd, plotdat, "-", color="#BEBEBE", linewidth=1.5)
            else:
                ax.plot(jdsub[twi], altaz[0][twi], "-",
                        color="#BEBEBE", linewidth=1.5)

        ax.plot(jdsub[night], altaz[0][night],
                "k", linewidth=1.5, label=plabel)
        ax.plot(jdsub[day], altaz[0][day], color="#FDB813", linewidth=1.5)

        altmax = np.argmax(altaz[0])
        ax.text(
            jdsub[altmax],
            altaz[0][altmax],
            str(n),
            color="b",
            fontsize=14,
            fontproperties=font1,
            va="bottom",
            ha="center",
        )

        if n == 29:
            ax.text(
                1.1,
                1.0 - float(n) * 0.04,
                "too many transits",
                ha="left",
                va="top",
                transform=ax.transAxes,
                fontsize=10,
                fontproperties=font0,
                color="r",
            )
        else:
            ax.text(
                1.1,
                1.0 - float(n) * 0.04,
                plabel,
                ha="left",
                va="top",
                transform=ax.transAxes,
                fontsize=12,
                fontproperties=font0,
                color="b",
            )

    ax.text(
        1.1,
        1.03,
        "Start of observation",
        ha="left",
        va="top",
        transform=ax.transAxes,
        fontsize=12,
        fontproperties=font0,
        color="b",
    )
    ax.text(
        1.1,
        1.0,
        "[No.]  Date",
        ha="left",
        va="top",
        transform=ax.transAxes,
        fontsize=12,
        fontproperties=font0,
        color="b",
    )

    axrange = ax.get_xlim()
    ax.set_xlabel("UT [hours]")

    if axrange[1] - axrange[0] <= 1.0:
        jdhours = np.arange(0, 3, 1.0 / 24.0)
        utchours = (np.arange(0, 72, dtype=int) + 12) % 24
    else:
        jdhours = np.arange(0, 3, 1.0 / 12.0)
        utchours = (np.arange(0, 72, 2, dtype=int) + 12) % 24
    ax.set_xticks(jdhours)
    ax.set_xlim(axrange)
    ax.set_xticklabels(utchours, fontsize=18)

    # Make ax2 responsible for "top" axis and "right" axis
    ax2 = ax.twin()
    # Set upper x ticks
    ax2.set_xticks(jdhours)
    ax2.set_xticklabels(utchours, fontsize=18)
    ax2.set_xlabel("UT [hours]")

    # Horizon angle for airmass
    airmass_ang = np.arange(5.0, 90.0, 5.0)
    geo_airmass = airmass.airmassPP(90.0 - airmass_ang)
    ax2.set_yticks(airmass_ang)
    airmassformat = []
    for t in range(geo_airmass.size):
        airmassformat.append("%2.2f" % geo_airmass[t])
    ax2.set_yticklabels(airmassformat, rotation=90)
    ax2.set_ylabel("Relative airmass", labelpad=32)
    ax2.tick_params(axis="y", pad=10, labelsize=10)
    plt.text(
        1.015,
        -0.04,
        "Plane-parallel",
        transform=ax.transAxes,
        ha="left",
        va="top",
        fontsize=10,
        rotation=90,
    )

    ax22 = ax.twin()
    ax22.set_xticklabels([])
    ax22.set_frame_on(True)
    ax22.patch.set_visible(False)
    ax22.yaxis.set_ticks_position("right")
    ax22.yaxis.set_label_position("right")
    ax22.spines["right"].set_position(("outward", 25))
    ax22.spines["right"].set_color("k")
    ax22.spines["right"].set_visible(True)
    airmass2 = np.array(
        [
            airmass.airmassSpherical(90.0 - ang, allData[n]["Obs coord"][2])
            for ang in airmass_ang
        ]
    )
    ax22.set_yticks(airmass_ang)
    airmassformat = []
    for t in range(airmass2.size):
        airmassformat.append("%2.2f" % airmass2[t])
    ax22.set_yticklabels(airmassformat, rotation=90)
    ax22.tick_params(axis="y", pad=10, labelsize=10)
    plt.text(
        1.045,
        -0.04,
        "Spherical+Alt",
        transform=ax.transAxes,
        ha="left",
        va="top",
        fontsize=10,
        rotation=90,
    )

    ax3 = ax.twin()
    ax3.set_frame_on(True)
    ax3.patch.set_visible(False)
    ax3.xaxis.set_ticks_position("bottom")
    ax3.xaxis.set_label_position("bottom")
    ax3.spines["bottom"].set_position(("outward", 50))
    ax3.spines["bottom"].set_color("k")
    ax3.spines["bottom"].set_visible(True)

    ltime, ldiff = localtime.localTime(
        utchours, np.repeat(allData[n]["Obs coord"][0], len(utchours))
    )
    jdltime = jdhours - ldiff / 24.0
    ax3.set_xticks(jdltime)
    ax3.set_xticklabels(utchours)
    ax3.set_xlim([axrange[0], axrange[1]])
    ax3.set_xlabel("Local time [hours]")

    ax.yaxis.set_major_locator(MultipleLocator(15))
    ax.yaxis.set_minor_locator(MultipleLocator(5))

    # Major tick format and font size
    ax.yaxis.set_major_formatter('{x:.0f}'+r"$^\circ$")
    ax.tick_params(axis='y', which='major', labelsize=20)
    # Minor tick format (fontsize default)
    ax.yaxis.set_minor_formatter('{x:.0f}'+r"$^\circ$")

    ax.set_ylabel("Altitude", fontsize=18)

    ax.yaxis.grid(color="gray", linestyle="dashed")
    ax.yaxis.grid(color="gray", which="minor", linestyle="dotted")
    ax2.xaxis.grid(color="gray", linestyle="dotted")

    def decifnec(s):
        try:
            r = s.decode("utf8")
        except AttributeError:
            r = s
        return r

    plt.text(
        0.5,
        0.95,
        "Transit visibility of " + decifnec(allData[n]["Planet name"]),
        transform=fig.transFigure,
        ha="center",
        va="bottom",
        fontsize=20,
    )

    if plotLegend:
        line1 = matplotlib.lines.Line2D(
            (0, 0), (1, 1), color="#FDB813", linestyle="-", linewidth=2
        )
        line2 = matplotlib.lines.Line2D(
            (0, 0), (1, 1), color="#BEBEBE", linestyle="-", linewidth=2
        )
        line3 = matplotlib.lines.Line2D(
            (0, 0), (1, 1), color="k", linestyle="-", linewidth=2
        )
        line4 = matplotlib.lines.Line2D(
            (0, 0), (1, 1), color="g", linestyle="-", linewidth=6, alpha=0.3
        )

        if markTransit:
            lgd2 = plt.legend(
                (line1, line2, line3, line4),
                (
                    "day",
                    "twilight",
                    "night",
                    "transit",
                ),
                bbox_to_anchor=(0.88, 0.15),
                loc=2,
                borderaxespad=0.0,
                prop={"size": 12},
                fancybox=True,
            )
        else:
            lgd2 = plt.legend(
                (line1, line2, line3),
                (
                    "day",
                    "twilight",
                    "night",
                ),
                bbox_to_anchor=(0.88, 0.13),
                loc=2,
                borderaxespad=0.0,
                prop={"size": 12},
                fancybox=True,
            )
        lgd2.get_frame().set_alpha(0.5)

    targetco = r"Target coordinates: (%8.4f$^\circ$, %8.4f$^\circ$)" % (
        allData[n]["Star ra"],
        allData[n]["Star dec"],
    )
    obsco = r"Obs coord.: (%8.4f$^\circ$, %8.4f$^\circ$, %4d m)" % (
        allData[n]["Obs coord"][0],
        allData[n]["Obs coord"][1],
        allData[n]["Obs coord"][2],
    )
    plt.text(
        0.01,
        0.97,
        targetco,
        transform=fig.transFigure,
        ha="left",
        va="center",
        fontsize=10,
    )
    plt.text(
        0.01,
        0.95,
        obsco,
        transform=fig.transFigure,
        ha="left",
        va="center",
        fontsize=10,
    )

    if print2file == True:
        outfile = "transVis-" + \
            str(allData[n]["Planet name"]).replace(" ", "") + ".png"
        plt.savefig(outfile, format="png", dpi=300)
    elif isinstance(print2file, six.string_types):
        plt.savefig(print2file, format="png", dpi=300)
    else:
        plt.show()



from PyAstronomy.pyaC import ImportCheck

ic = ImportCheck(["numpy", "scipy"], required=["numpy"])

from .XTran import palTrans
from .XTran import forTrans
from .XTran import LimBrightTrans, RmcL, RmcL_Hirano, RmcLell
from .radVel import SinRadVel
from .planetBrightnessPhases import GeomPlanetBrightPhase
from .keplerEllipseModel import KeplerEllipseModel, KeplerRVModel
#from atanProfile import AtanProfile, AtanProfileDamped
from .lineListGaussModel import *
from .voigtAstro import VoigtAstroP
from .LyAProfile import LyaTransmission
from .rotBroadProfile import RotBroadProfile


# -*- coding: utf-8 -*-
import numpy as np

def flux2photons(wave, flux):
  """
    Convert flux (erg/s) to photons/s.
    
    Parameters
    ----------
    wave : float, array
        The wavelength in Angstrom.
    flux : float, array
        The associated flux array in erg/s.
   
    Returns
    -------
    photons : array
        Number of photons/s for the given input flux

  """
  # h*c = 1.98644568327e-08 erg*A; ephoton = energy(erg) per photon
  ePhoton = 1.98644568327e-08 / np.array(wave)
  nPhotons = np.array(flux) / ePhoton
  return nPhotons
  
def photons2flux(wave, photons):
  """
    Convert photons/s to flux (erg/s).
    
    Parameters
    ----------
    wave : float, array
        The wavelength in Angstrom.
    photons : float, array
        The associated photons array in photons/s.
   
    Returns
    -------
    flux : array
        Flux in erg/s for the given photon flux
  """
  # h*c = 1.98644568327e-08 erg*A; ephoton = energy(erg) per photon
  ePhoton = 1.98644568327e-08 / np.array(wave)
  flux = np.array(photons) * ePhoton
  return flux  


from __future__ import division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import pyasl


class _ZList:
    """
    Calculate the projected distance between the centers of star and planet.

    The resulting values will be stored in the `_zlist` attribute. Additionally,
    the indices belonging to in-transit points will be saved in the `_intrans`
    attribute. Indices pertaining to the occultation (secondary eclipse) are
    available via the `_inocc` attribute.
    These can be used as input to calculate the transit light-curve.

    Parameters
    ----------
    orbit : string
        Either "circular" or "keplerian".
    """

    def _zlistCirc(self, time):
        """
        Calculate projected distance in case of circular orbit.

        Required parameters: per, i, a (assumes transit center at 0)

        Parameters
        ----------
        time : array
            The time points (unit has to be consistent throughout all parameters
            used for calculation).
        """
        # w = circular frequency
        w = 2.0 * np.pi / self["per"]
        # The angle needed to realize the orbit inclination
        alpha = (90.0 - self["i"]) / 180.0 * np.pi
        # Determine the phases (to ensure that the secondary eclipse is not considered)
        phase = time / self["per"]
        phase -= np.floor(phase)

        # Distance between centers
        wt = w * time
        self._zlist = self["a"] * np.sqrt(
            np.sin(wt) ** 2 + np.sin(alpha) ** 2 * np.cos(wt) ** 2
        )

        # In these cases the 'planet' will be in front of the primary
        self._intrans = np.where(
            (np.abs(self._zlist) < (1 + self["p"]))
            & np.logical_or(phase > 0.75, phase < 0.25)
        )[0]

        # Here, the planet is behind the star (occultation)
        self._inocc = np.where(
            (np.abs(self._zlist) < (1 + self["p"]))
            & np.logical_and(phase > 0.25, phase < 0.75)
        )[0]

    def _zlistKep(self, time):
        """
        Calculate the projected, normalized distance of star and planet centers.

        Required parameters: per, i, a, e, w, T0pa, Omega, p
        Additionally requires: collisionCheck

        Note that in this method the line of sight is the -z direction.

        Parameters
        ----------
        time : array
            The time points (unit has to be consistent throughout all
            parameters used for calculation).
        """
        # Assign values to KeplerEllipse
        self._ke.a = self["a"]
        self._ke.e = self["e"]
        self._ke.i = self["i"]
        self._ke.Omega = self["Omega"]
        self._ke.per = self["per"]
        self._ke.tau = self["tau"]
        self._ke.w = self["w"]

        # Initialize the z-list (NaN values will be neglected)
        self._zlist = np.empty(len(time))
        self._zlist[:] = np.nan
        # Get position of body from Keplerian orbit
        pos = self._ke.xyzPos(time)

        # Default line of sight is in +z direction (i.e., observer
        # located at -z)

        # Calculate projected distance between centers of star and
        # planet sqrt(x**2 + y**2)
        z = np.sqrt(pos[::, 0] ** 2 + pos[::, 1] ** 2)

        # If enabled, check for body collisions
        if self._collisionCheck:
            r = np.sqrt(pos[::, 0] ** 2 + pos[::, 1] ** 2 + pos[::, 2] ** 2)
            indi = np.where(r < 1.0 + self["p"])[0]
            if len(indi) > 0:
                raise (
                    PE.PyAValError(
                        "There is a body collision on this orbit.", where="_ZList"
                    )
                )

        # Determine which values are relevant. In particular, these are those
        # for which the planet is in front of the star (z > 0) and the distance
        # between stellar and planetary distance is lower than 1+p.
        self._intrans = np.where(
            np.logical_and(z <= (1.0 + self["p"]), pos[::, 2] < 0.0)
        )[0]
        self._inocc = np.where(
            np.logical_and(z <= (1.0 + self["p"]), pos[::, 2] > 0.0)
        )[0]

        # Calculate the 'orbit' and store z-values in _zlist
        self._zlist = z

    def __init__(self, orbit, cc=True):
        if orbit == "circular":
            self._calcZList = self._zlistCirc
        elif orbit == "keplerian":
            self._collisionCheck = cc
            self._ke = pyasl.KeplerEllipse(1.0, 1.0)
            self._calcZList = self._zlistKep



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from . import periodBase
import numpy
import scipy.special


class Fourier(periodBase.PeriodBase):
    r"""
    This class computes the Fast Fourier Transform (FFT) of the input data.

    This is actually a wrapper around numpy's FFT routines.
    The constructor takes the light
    curve, `lc` (*TimeSeries* instance), as input. The optional
    argument specifies the normalization of the Fourier Power.
    Currently, only the normalization according to [Leahy83]_
    is supported, which in the case of purely Poissonian
    noise results in a mean power of 2.

    Parameters
    ----------
    lc : TimeSeries instance
        The light curve to be analyzed.
    norm : optional, string
        Normalization method; currently, only default ("Leahy") is supported.
    """

    def __calcPeriodogram(self):

        dt = self.t[1:]-self.t[0:-1]
        if numpy.max(dt) > 1.1 * numpy.min(dt):
            print("timingAna::fourier() - Error: data must be equally spaced!")
            exit()

        power = numpy.abs(numpy.fft.fft(self.y))**2.
        if self.norm == "Leahy":
            power = power*2./numpy.sum(self.y)
        else:
            print("Normalization not yet implemented!")
            power = power*2./numpy.sum(self.y)
        freq = numpy.fft.fftfreq(numpy.size(self.y), numpy.mean(dt))
        indi = numpy.where(freq > 0)
        self.freq = freq[indi]
        self.power = power[indi]
        self.M = len(self.freq)

    def __init__(self, lc, norm="Leahy"):
        self.freq = None
        self.power = None
        self.t = lc.time
        self.y = lc.flux
        self.norm = norm
        self.label = {'title': 'Fourier Transformation',
                      'xlabel': 'Frequency',
                      'ylabel': 'Fourier Power'}
        self.__calcPeriodogram()

    def Prob(self, Pn):
        r"""
        Returns the probability to obtain a power larger than the threshold, `Pn`.

        In the calculations, the noise is assumed to be Poisson-distributed.

        .. note::
            According to [vdK]_ the probability \
            to obtain a power larger than a given threshold from the noise is given by

            .. math::
              Prob(p>Pn) = Q(M \times W \times Pn, 2 \times M \times W)

            where :math:`Q(\chi^2, \nu)` is the cumulative :math:`\chi^2` distribution with :math:`\nu` d.o.f.

        Parameters
        ----------
        Pn : float
            Power threshold.

        Returns
        -------
        FAP : float
            The probability to obtain a power larger than the specified
            threshold from noise.
        """
        nu = 2
        # -- Integral from x to infinity of Chi-square pdf.
        Q = scipy.special.chdtrc(nu, Pn)
        return Q



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo

def expCorrRN(n, tau, mean=0.0, std=1.0, rnos=None, fullOut=False):
  """
    Generate exponentially correlated random numbers.
    
    This procedure implements the prescription given by
    Deserno 2002 ("How to generate exponentially correlated Gaussian
    random numbers"). The autocorrelation function of the resulting
    numbers decays with the predefined "decay time", tau.
    The correlation coefficient of the resulting numbers is given
    by exp(-1/tau).
    
    Parameters
    ----------
    n : int
        Number of numbers to be generated.
    tau : float
        Decay time
    mean : float, optional
        Mean of the numbers to be generated.
        Default is 0.0.
    std : float, optional
        Standard deviation of the generated numbers.
        Default is 1.0.
    rnos : array, optional
        Uncorrelated Gaussian random numbers with mean 0.0
        and standard deviation 1.0 to be used to generate
        correlated random numbers. If not given, Gaussian
        random numbers will be obtained using numpy.random.normal.
    fullOut : boolean, optional
        If False (default), only the correlated random numbers
        will be returned. 
    
    Returns
    -------
    Correlated RNs : array
        Correlated Gaussian random numbers.
    Uncorrelated RNs : array, optional
        The uncorrelated random numbers used to generate
        the correlated numbers (only of `fullOut` is True).
    Correlation coefficient : float
        The correlation coefficient (exp(-1/tau), only of
        `fullOut` is True).
  """
  if rnos is None:  
    # Get n uniformly distributed random numbers between 0 and 1
    g = np.random.normal(0.0, 1.0, n)
  else:
    g = rnos
  
  if len(g) != n:
    raise(PE.PyAValError("The length of `rnos` must be n.",
                         where="expCorrRN",
                         solution=["Adjust `rnos`.", \
                                   "Remove rnos argument to use internally generated Gaussian random numbers."]))
    
  # Prepare result
  result = np.zeros(n)
  result[0] = g[0]
  
  # Correlation coefficient
  f = np.exp(-1.0/tau)
  
  # Obtain correlated numbers
  somf = np.sqrt(1.0 - f**2)
  for i in smo.range(1,n):
    result[i] = f*result[i-1] + somf*g[i]
  result = mean + std*result
  
  if fullOut:
    return result, g, f
  return result


import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE

def getCardinalPoint(azimuth):
  """
    Get the cardinal point (North, East, South, or West) for a given azimuth.
    
    Here, the azimuth is measured from North to East.
    N = (315, 45] deg
    E = (45, 135] deg
    S = (135, 225] deg
    W = (225, 315] deg

    Parameters
    ----------
    azimuth : float
        Azimuth of an object in deg (0-360).

    Returns
    -------
    Cardinal point : string, {N,E,S,W}
        Returns the cardinal point (N, E, S, W) corresponding
        to the given azimuth of the object.
  """  
  
  if (azimuth < 0.0) or (azimuth > 360.):
    raise(PE.PyAValError("The azimuth needs to be a number in the range [0,360]. Given azimuth: " + str(azimuth), \
          where="getCardinalPoint", \
          solution="Specify a number between 0 and 360"))
  
  if np.logical_and(azimuth > 315., azimuth <= 360.): return "N"
  if np.logical_and(azimuth >= 0., azimuth <= 45.):   return "N"
  if np.logical_and(azimuth > 45., azimuth <= 135.):  return "E"
  if np.logical_and(azimuth > 135., azimuth <= 225.): return "S"
  if np.logical_and(azimuth > 225., azimuth <= 315.): return "W"


# -*- coding: utf-8 -*-

# Copyright (c) 2012 Sebastian Schröter, Stefan Czesla, and Mathias Zechmeister

# The MIT License (MIT)

# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division
from . import periodBase
from numpy import sum, cos, sin, arange, min, max, pi, \
                  exp, log, zeros, argmax, sqrt, arctan2, dot
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE 



class Gls:
    """
    Compute the Generalized Lomb-Scargle (GLS) periodogram.

    The *Gls* class computes the error-weighted Lomb-Scargle periodogram as
    developed by [ZK09]_ using various possible normalizations.

    The constructor of *Gls* takes a *TimeSeries* instance (i.e., a light curve)
    as first argument. The constructor allows to pass keywords to adjust the
    `freq` array, which will be used to calculate the periodogram.

    The main result of the calculation, i.e., the power, are stored in the
    class property `power`.

    Parameters
    ----------
    lc : tuple or list or TimeSeries object
        The light curve data either in the form of a TimeSeries object (or any
        object providing the attributes time, flux, and error) or a tuple or list
        providing time as first element, flux as second element, and optionally,
        the error as third element.
    fbeg, fend : float, optional
        The beginning and end frequencies for the periodogram
        (inverse units of time axis).
    Pbeg, Pend : float, optional
        The beginning and end periods for the periodogram
        (same units as for time axis).
    ofac : int
        Oversampling factor of frequency grid (default=10).
    hifac : float
        Maximum frequency `freq` = `hifac` * (average Nyquist frequency)
        (default=1).
    freq : array, optional
        Contains the frequencies at which to calculate the periodogram.
        If given, fast and verbose option are not available.
        If not given, a frequency array will be automatically generated.
    norm : string, optional
        The normalization; either of "ZK", "Scargle", "HorneBaliunas", "Cumming", "wrms", "chisq".
        The default is unity ("ZK").
    ls : boolean, optional
        If True, the conventional Lomb-Scargle periodogram will be computed
        (default is False).
    fast : boolean, optional
        If True, recursive relations for trigonometric functions will be used
        leading to faster evaluation (default is False).
    verbose : boolean, optional
        Set True to obtain some statistical output (default is False).

    Attributes
    ----------
    power : array
        The normalized power of the GLS.
    freq : array
        The frequency array.
    ofac : int
        The oversampling factor of frequency grid.
    hifac : float
        The maximum frequency.
    t : array
        The abscissa data values.
    y : array
        The ordinate data values.
    e_y : array
        The errors of the data values.
    norm : string, {'ZK', 'Scargle', 'HorneBaliunas', 'Cumming', 'wrms', 'chisq'}
        The used normalization.

    Examples
    --------
    Create 1000 unevenly sampled data points with frequency=0.1,
    measurement error and Gaussian noise
    
    >>> time = np.random.uniform(54000., 56000., 1000)
    >>> flux = 0.15 * np.sin(2. * np.pi * time / 10.)

    Add some noise
    
    >>> error = 0.5 * np.ones(time.size)
    >>> flux += np.random.normal(0, error)

    Compute the full error-weighted Lomb-Periodogram
    in 'ZK' normalization and calculate the significance
    of the maximum peak.
    
    >>> gls = Gls((time, flux, error), verbose=True)

    >>> maxPower = gls.pmax
    >>> print("GLS maximum power: ", maxPower)
    >>> print("GLS statistics of maximum power peak: ", gls.stats(maxPower))
    >>> gls.plot(block=True)

    """
    # Available normalizations
    norms = ['ZK', 'Scargle', 'HorneBaliunas', 'Cumming', 'wrms', 'chisq', 'lnL', 'dlnL']

    def __init__(self, lc, fbeg=None, fend=None, Pbeg=None, Pend=None, ofac=10, hifac=1, freq=None, norm="ZK", ls=False, fast=False, verbose=False, **kwargs):

        self.freq = freq
        self.fbeg = fbeg
        self.fend = fend
        self.Pbeg = Pbeg
        self.Pend = Pend
        self.ofac = ofac
        self.hifac = hifac
        self.ls = ls
        self.norm = norm
        self.fast = fast
        self.label = {'title': 'Generalized Lomb Periodogram',
                      'xlabel': 'Frequency'}
        if "stats" in kwargs:
            print("Warning: 'stats' option is outdated. Please use 'verbose' instead.")
            verbose = kwargs["stats"]

        self._normcheck(norm)

        self._assignTimeSeries(lc)
        self._buildFreq()
        self._calcPeriodogram()
        self.pnorm(norm)
        self._peakPeriodogram()

        # Output statistics
        if verbose:
            self.info()

    def _sete_y(self, ey):
        self.yerr = ey
    
    def _gete_y(self):
        return self.yerr

    """ Treat e_y and yerr synonymously for backward compatibility """
    e_y = property(_gete_y, _sete_y)
    

    def _assignTimeSeries(self, lc):
        """
        A container class that holds the observed light curve.

        Parameters
        ----------
        time : array
            The time array.
        flux : array
            The observed flux/data.
        error : array, optional
            The error of the data values.

        """
        def checkvalid(x, m):
            # Check for NaNs
            if np.any(~np.isfinite(x)):
                raise(PE.PyAValError(m + " contains invalid values such as NaN or Inf.", \
                                    where="GLS", \
                                    solution="Remove from input (e.g., using np.isfinite)"))                    
        
        if isinstance(lc, (tuple, list)):
            # t, y[, e_y] were given as list or tuple.
            if len(lc) in (2, 3):
                self.t = np.ravel(lc[0])
                self.y = np.ravel(lc[1])
                # Check for NaNs
                checkvalid(self.t, "Time input")
                checkvalid(self.y, "Flux input")
                self.e_y = None
                if (len(lc) == 3) and (lc[2] is not None):
                    # Error has been specified.
                    self.e_y = np.ravel(lc[2])
                    checkvalid(self.e_y, "Flux error input")
            else:
                raise(ValueError("lc is a list or tuple with " + str(len(lc)) + " elements. Needs to have 2 or 3 elements." + \
                                   " solution=Use 2 or 3 elements (t, y[, e_y]) or an instance of TimeSeries"))
        else:
            # Assume lc is an instance of TimeSeries.
            self.t, self.y, self.e_y = lc.time, lc.flux, lc.error

        self.th = self.t - self.t.min()
        self.tbase = self.th.max()
        self.N = len(self.y)

        # Re-check array length compatibility
        if (len(self.th) != self.N) or ((self.e_y is not None) and (len(self.e_y) != self.N)):
            raise(ValueError("Incompatible dimensions of input data arrays (time and flux [and error]). Current shapes are: " + \
                             ', '.join(str(np.shape(x)) for x in (self.t, self.y, self.e_y))))
            

    def _buildFreq(self):
        """
        Build frequency array (`freq` attribute).

        Attributes
        ----------
        fnyq : float
            Half of the average sampling frequency of the time series.

        """
        self.fstep = 1 / self.tbase / self.ofac   # frequency sampling depends on the time span, default for start frequency
        self.fnyq = 0.5 / self.tbase * self.N     # Nyquist frequency
        self.f = self.freq

        if self.freq is None:
            # Build frequency array if not present.
            if self.fbeg is None:
                self.fbeg = self.fstep if self.Pend is None else 1 / self.Pend
            if self.fend is None:
                self.fend = self.fnyq * self.hifac if self.Pbeg is None else 1 / self.Pbeg

            if self.fend <= self.fbeg:
                raise(ValueError("fend is smaller than (or equal to) fbeg but it must be larger." + \
                               "Choose fbeg and fend so that fend > fbeg."))

            self.freq = arange(self.fbeg, self.fend, self.fstep)
        else:
            if self.fast:
                raise(ValueError("freq and fast cannot be used together."))
            self.fbeg, self.fend = np.min(self.freq), np.max(self.freq)

        self.nf = len(self.freq)

        # An ad-hoc estimate of the number of independent frequencies (Eq. (24) in ZK_09).
        self.M = (self.fend-self.fbeg) * self.tbase

    def _calcPeriodogram(self):

        if self.e_y is None:
            w = np.ones(self.N)
        else:
            w = 1 / (self.e_y * self.e_y)
        self.wsum = w.sum()
        w /= self.wsum

        self._Y = dot(w, self.y)       # Eq. (7)
        wy = self.y - self._Y          # Subtract weighted mean
        self._YY = dot(w, wy**2)       # Eq. (10)
        wy *= w                        # attach errors

        C, S, YC, YS, CC, CS = np.zeros((6, self.nf))

        if self.fast:
            # Prepare trigonometric recurrences.
            eid = exp(2j * pi * self.fstep * self.th)  # cos(dx)+i sin(dx)

        for k, omega in enumerate(2.*pi*self.freq):
            # Circular frequencies.
            if self.fast:
                if k % 1000 == 0:
                    # init/refresh recurrences to stop error propagation
                    eix = exp(1j * omega * self.th)  # exp(ix) = cos(x) + i*sin(x)
                cosx = eix.real
                sinx = eix.imag
                eix *= eid              # increase freq for next loop
            else:
                x = omega * self.th
                cosx = cos(x)
                sinx = sin(x)

            C[k] = dot(w, cosx)         # Eq. (8)
            S[k] = dot(w, sinx)         # Eq. (9)

            YC[k] = dot(wy, cosx)       # Eq. (11)
            YS[k] = dot(wy, sinx)       # Eq. (12)
            wcosx = w * cosx
            CC[k] = dot(wcosx, cosx)    # Eq. (13)
            CS[k] = dot(wcosx, sinx)    # Eq. (15)

        SS = 1. - CC
        if not self.ls:
            CC -= C * C            # Eq. (13)
            SS -= S * S            # Eq. (14)
            CS -= C * S            # Eq. (15)
        D = CC*SS - CS*CS          # Eq. (6)

        self._a = (YC*SS-YS*CS) / D
        self._b = (YS*CC-YC*CS) / D
        self._off = -self._a*C - self._b*S

        # power
        self.p = (SS*YC*YC + CC*YS*YS - 2.*CS*YC*YS) / (self._YY*D)   # Eq. (5) in ZK09

    def _normcheck(self, norm):
        """
        Check normalization

        Parameters
        ----------
        norm : string
            Normalization string

        """
        if norm not in self.norms:
            raise(ValueError("Unknown norm: " + str(norm) + ". " + \
                "Use either of " + ', '.join(self.norms)))

    def pnorm(self, norm="ZK"):
        """
        Assign or modify normalization (can be done afterwards).

        Parameters
        ----------
        norm : string, optional
            The normalization to be used (default is 'ZK').

        Examples
        --------
        >>> gls.pnorm('wrms')

        """
        self._normcheck(norm)
        self.norm = norm
        p = self.p
        power = p   # default ZK
        self.label["ylabel"] = "Power ("+norm+")"

        if norm == "Scargle":
            popvar = input('pyTiming::gls - Input a priori known population variance:')
            power = p / float(popvar)
        elif norm == "HorneBaliunas":
            power = (self.N-1)/2. * p
        elif norm == "Cumming":
            power = (self.N-3)/2. * p / (1.-self.p.max())
        elif norm == "chisq":
            power = self._YY *self.wsum * (1.-p)
            self.label["ylabel"] = "chisq"
        elif norm == "wrms":
            power = sqrt(self._YY*(1.-p))
            self.label["ylabel"] = "wrms"
        elif norm == "lnL":
            chi2 = self._YY *self.wsum * (1.-p)
            power = -0.5*chi2 - 0.5*np.sum(np.log(2*np.pi * self.e_y))
            self.label["ylabel"] = "lnL"
        elif norm == "dlnL":
            # dlnL = lnL - lnL0 = -0.5 chi^2 + 0.5 chi0^2 = 0.5 (chi0^2 - chi^2) = 0.5 chi0^2 p
            power = 0.5 * self._YY * self.wsum * p
            self.label["ylabel"] = "dlnL"

        self.power = power

    def _peakPeriodogram(self):
        """
        Analyze the highest periodogram peak.
        """
        # Index with maximum power
        k = argmax(self.p)
        # Maximum power
        self.pmax = pmax = self.p[k]
        self.rms = rms = sqrt(self._YY*(1.-pmax))

        # Statistics of highest peak
        self.hpstat = p = {}

        # Best parameters
        p["fbest"] = fbest = self.freq[k]
        p["amp"] = amp = sqrt(self._a[k]**2 + self._b[k]**2)
        p["ph"] = ph = arctan2(self._a[k], self._b[k]) / (2.*pi)
        p["T0"]  = self.t.min() - ph/fbest
        p["offset"] = self._off[k] + self._Y            # Re-add the mean.

        # Error estimates
        p["amp_err"] = sqrt(2./self.N) * rms
        p["ph_err"] = ph_err = sqrt(2./self.N) * rms/amp/(2.*pi)
        p["T0_err"] = ph_err / fbest
        p["offset_err"] = sqrt(1./self.N) * rms

        # Get the curvature in the power peak by fitting a parabola y=aa*x^2
        if 1 < k < self.nf-2:
            # Shift the parabola origin to power peak
            xh = (self.freq[k-1:k+2] - self.freq[k])**2
            yh = self.p[k-1:k+2] - pmax
            # Calculate the curvature (final equation from least square)
            aa = dot(yh, xh) / dot(xh, xh)
            p["f_err"] = e_f = sqrt(-2./self.N / aa * (1.-self.pmax))
            p["Psin_err"] = e_f / fbest**2
        else:
            self.hpstat["f_err"] = np.nan
            self.hpstat["Psin_err"] = np.nan
            print("WARNING: Highest peak is at the edge of the frequency range.\nNo output of frequency error.\nIncrease frequency range to sample the peak maximum.")

    def sinmod(self, t):
        """
        Calculate best-fit sine curve.

        The parameters of the best-fit sine curve can be accessed via
        the dictionary attribute `hpstat`. Specifically, "amp" holds the
        amplitude, "fbest" the best-fit frequency, "T0" the reference time
        (i.e., time of zero phase), and "offset" holds the additive offset
        of the sine wave. 

        Parameters
        ----------
        t : array
            Time array at which to calculate the sine.

        Returns
        -------
        Sine curve : array
            The best-fit sine curve (i.e., that for which the
            power is maximal).
        """
        try:
            p = self.hpstat
            return p["amp"] * sin(2*np.pi*p["fbest"]*(t-p["T0"])) + p["offset"]
        except Exception as e:
            print("Failed to calculate best-fit sine curve.")
            raise(e)

    def info(self, noprint=False):
        """
        Prints some basic statistical output screen.
        
        Parameters
        ----------
        noprint : boolean
            If True, printing is suppressed. Default is False.
        
        Returns
        -------
        Information : dictionary
            A dictionary with the printed information.
        """
        
        info = {"N":self.N, "weighted_mean":self._Y, "weighted_rms":sqrt(self._YY), \
                "time_base":self.tbase, "no_of_freq_points":self.nf, \
                "norm":self.norm, "maximum_power":self.power.max(), \
                "rms_of_residuals":self.rms, \
                "best_sine_frequency":self.hpstat["fbest"], \
                "best_sine_frequency_err":self.hpstat["f_err"], \
                "best_sine_period":1./self.hpstat["fbest"], \
                "best_sine_period_err":self.hpstat["Psin_err"], \
                "amplitude":self.hpstat["amp"], \
                "amplitude_err":self.hpstat["amp_err"], \
                "phase":self.hpstat["ph"], \
                "phase_err":self.hpstat["ph_err"], \
                "phase_t0":self.hpstat["T0"], \
                "phase_t0_err":self.hpstat["T0_err"], \
                "offset":self.hpstat["offset"], \
                "offset_err":self.hpstat["offset_err"]
            }
        if self.e_y is not None:
            info["mean_weighted_internal_error"] = sqrt(self.N/sum(1./self.e_y**2)),
        
        if not noprint:
            print("Generalized LS - statistical output")
            print("-----------------------------------")
            print("Number of input points:      %-6d" % self.N)
            print("Weighted mean of dataset:   % f"  % self._Y)
            print("Weighted rms of dataset:    % f"  % sqrt(self._YY))
            print("Time base:                  % f"  % self.tbase)
            print("Number of frequency points:  %-6d" % self.nf)
            print()
            print("Maximum power p [%s]: % f" % (self.norm, self.power.max()))
            print("RMS of residuals:     % f" % self.rms)
            if self.e_y is not None:
                print("  Mean weighted internal error:  %f" % (sqrt(self.N/sum(1./self.e_y**2))))
            print("Best sine frequency:  % f +/- % f" % (self.hpstat["fbest"], self.hpstat["f_err"]))
            print("Best sine period:     % f +/- % f" % (1./self.hpstat["fbest"], self.hpstat["Psin_err"]))
            print("Amplitude:            % f +/- % f" % (self.hpstat["amp"], self.hpstat["amp_err"]))
            print("Phase (ph):           % f +/- % f" % (self.hpstat["ph"], self.hpstat["ph_err"]))
            print("Phase (T0):           % f +/- % f" % (self.hpstat["T0"], self.hpstat["T0_err"]))
            print("Offset:               % f +/- % f" % (self.hpstat["offset"], self.hpstat["offset_err"]))
            print("-----------------------------------")
        return info

    def plot(self, block=False, period=False):
        """
        Create a plot.
        """
        try:
            import matplotlib
            if (matplotlib.get_backend() != "TkAgg"):
                matplotlib.use("TkAgg")
            import matplotlib.pylab as plt
            from matplotlib.ticker import FormatStrFormatter
        except ImportError:
            raise(ImportError("Could not import matplotlib.pylab."))

        fig = plt.figure()
        fig.subplots_adjust(hspace=0.15, wspace=0.08, right=0.97, top=0.95)
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Normalized periodogram")
        if period:
            ax.set_xscale("log")
            ax.set_xlabel("Period")
        else:
            ax.set_xlabel("Frequency")
        ax.set_ylabel(self.label["ylabel"])
        ax.plot(1/self.freq if period else self.freq, self.power, 'b-')

        fbest, T0 = self.hpstat["fbest"], self.hpstat["T0"]
        # Data and model
        datstyle = {'yerr':self.e_y, 'fmt':'r.', 'capsize':0}
        tt = arange(self.t.min(), self.t.max(), 0.01/fbest)
        ymod = self.sinmod(tt)
        yfit = self.sinmod(self.t)
        ax1 = fig.add_subplot(3, 2, 3)
        # ax1.set_xlabel("Time")
        ax1.set_ylabel("Data")
        plt.setp(ax1.get_xticklabels(), visible=False)
        ax1.errorbar(self.t, self.y, **datstyle)
        ax1.plot(tt, ymod, 'b-')

        tt = arange(T0, T0+1/fbest, 0.01/fbest)
        yy = self.sinmod(tt)
        ax2 = fig.add_subplot(3, 2, 4, sharey=ax1)
        plt.setp(ax2.get_xticklabels(), visible=False)
        plt.setp(ax2.get_yticklabels(), visible=False)
        # ax2.set_xlabel("Time")
        # ax2.set_ylabel("Data")
        ax2.errorbar(self.t*fbest % 1, self.y, **datstyle)
        xx = tt*fbest % 1
        ii = np.argsort(xx)
        ax2.plot(xx[ii], yy[ii], 'b-')

        # Residuals
        yres = self.y - yfit
        ax3 = fig.add_subplot(3, 2, 5, sharex=ax1)
        ax3.set_xlabel("Time")
        ax3.set_ylabel("Residuals")
        ax3.errorbar(self.t, yres, **datstyle)
        ax3.plot([self.t.min(), self.t.max()], [0,0], 'b-')

        ax4 = fig.add_subplot(3, 2, 6, sharex=ax2, sharey=ax3)
        # ax4.set_title("Data")
        ax4.set_xlabel("Phase")
        # ax4.set_ylabel("Data")
        plt.setp(ax4.get_yticklabels(), visible=False)
        ax4.errorbar(self.t*fbest % 1, yres, **datstyle)
        ax4.plot([0,1], [0,0], 'b-')

        if hasattr(plt.get_current_fig_manager(), 'toolbar'):
            # check seems not needed when "TkAgg" is set
            plt.get_current_fig_manager().toolbar.pan()
        #t = fig.canvas.toolbar
        #plt.ToggleTool(plt.wx_ids['Pan'], False)
        if block:
            print("Close the plot to continue.")
        else:
            plt.ion()
        plt.show()
        # plt.show(block=block) # unexpected keyword argument 'block' in older matplotlib
        return plt

    def prob(self, Pn):
        """
        Probability of obtaining the given power.

        Calculate the probability to obtain a power higher than
        `Pn` from the noise, which is assumed to be Gaussian.

        .. note:: Normalization
          (see [ZK09]_ for further details).

          - `Scargle`:
          .. math::
            
             exp(-Pn)

          - `HorneBaliunas`:
          .. math::
              
             \\left(1 - 2 \\times \\frac{Pn}{N-1} \\right)^{(N-3)/2}

          - `Cumming`:
          .. math::
              
             \\left(1+2\\times \\frac{Pn}{N-3}\\right)^{-(N-3)/2}

        Parameters
        ----------
        Pn : float
            Power threshold.

        Returns
        -------
        Probability : float
            The probability to obtain a power equal or
            higher than the threshold from the noise.

        """
        self._normcheck(self.norm)
        if self.norm == "ZK": return (1.-Pn)**((self.N-3.)/2.)
        if self.norm == "Scargle": return exp(-Pn)
        if self.norm == "HorneBaliunas": return (1-2*Pn/(self.N-1)) ** ((self.N-3)/2)
        if self.norm == "Cumming": return (1+2*Pn/(self.N-3)) ** (-(self.N-3)/2)
        if self.norm == "wrms": return (Pn**2/self._YY) ** ((self.N-3)/2)
        if self.norm == "chisq": return (Pn/self._YY/self.wsum) ** ((self.N-3)/2)
        if self.norm == "ZK":
            p = Pn
        if self.norm == "dlnL":
            p = 2 * Pn / self._YY / self.wsum
        if self.norm == "lnL":
            chi2 = -2*Pn - np.sum(np.log(2*np.pi * self.e_y**2))
            p = 1 - chi2/self._YY/self.wsum
        return (1-p) ** ((self.N-3)/2)


    def probInv(self, Prob):
        """
        Calculate minimum power for given probability.

        This function is the inverse of `Prob(Pn)`.
        Returns the minimum power for a given probability threshold `Prob`.

        Parameters
        ----------
        Prob : float
            Probability threshold.

        Returns
        -------
        Power threshold : float
            The minimum power for the given false-alarm probability threshold.

        """
        self._normcheck(self.norm)
        if self.norm == "ZK": return 1.-Prob**(2./(self.N-3.))
        if self.norm == "Scargle": return -log(Prob)
        if self.norm == "HorneBaliunas": return (self.N-1) / 2 * (1-Prob**(2/(self.N-3)))
        if self.norm == "Cumming": return (self.N-3) / 2 * (Prob**(-2./(self.N-3))-1)
        if self.norm == "wrms": return sqrt(self._YY * Prob**(2/(self.N-3)))
        if self.norm == "chisq": return self._YY * self.wsum * Prob**(2/(self.N-3))
        p = 1 - Prob**(2/(self.N-3))
        if self.norm == "ZK": return p
        if self.norm == "lnL": return -0.5*self._YY*self.wsum*(1.-p) - 0.5*np.sum(np.log(2*np.pi * self.e_y**2))
        if self.norm == "dlnL": return 0.5 * self._YY * self.wsum * p


    def FAP(self, Pn):
        """
        Obtain the false-alarm probability (FAP).

        The FAP denotes the probability that at least one out of M independent
        power values in a prescribed search band of a power spectrum computed
        from a white-noise time series is as large as or larger than the
        threshold, `Pn`. It is assessed through

        .. math:: FAP(Pn) = 1 - (1-Prob(P>Pn))^M \\; ,

        where "Prob(P>Pn)" depends on the type of periodogram and normalization
        and is calculated by using the *prob* method; *M* is the number of
        independent power values and is computed internally.

        Parameters
        ----------
        Pn : float
            Power threshold.

        Returns
        -------
        FAP : float
            False alarm probability.

        """
        prob = self.M * self.prob(Pn)
        if prob > 0.01:
            return 1. - (1.-self.prob(Pn))**self.M
        return prob

    def powerLevel(self, FAPlevel):
        """
        Power threshold for FAP level.

        Parameters
        ----------
        FAPlevel : float or array_like
              "False Alarm Probability" threshold

        Returns
        -------
        Threshold : float or array
            The power threshold pertaining to a specified false-alarm
            probability (FAP). Powers exceeding this threshold have FAPs
            smaller than FAPlevel.

        """
        Prob = 1. - (1.-FAPlevel)**(1./self.M)
        return self.probInv(Prob)

    def stats(self, Pn):
        """
        Obtain basic statistics for power threshold.

        Parameters
        ----------
        Pn : float
            Power threshold.

        Returns
        -------
        Statistics : dictionary
            A dictionary containing {'Pn': *Pn*, 'Prob': *Prob(Pn)* ,
            'FAP': *FAP(Pn)*} for the specified power threshold, *Pn*.

        """
        return {'Pn': Pn, 'Prob': self.prob(Pn), 'FAP': self.FAP(Pn)}

    def toFile(self, ofile, header=True):
        """
        Write periodogram to file.

        The output file is a standard text file with two columns,
        viz., frequency and power. 

        Parameters
        ----------
        ofile : string
            Name of the output file.

        """
        with open(ofile, 'w') as f:
            if header:
                f.write("# Generalized Lomb-Scargle periodogram\n")
                f.write("# Parameters:\n")
                f.write("#    Data file: %s\n" % self.df)
                f.write("#    ofac     : %s\n" % self.ofac)
                f.write("#    norm     : %s\n" % self.norm)
                f.write("# 1) Frequency, 2) Normalized power\n")
            for line in zip(self.freq, self.power):
                f.write("%f  %f\n" % line)

        print("Results have been written to file: ", ofile)



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
import os
from .. import readFitsSpec
import numpy as np
import astropy.io.fits as pyfits


class SpectralLib:
    
    def refreshDescription(self):
        """
        (Re-)Download file description.txt
        
        This file contains an identifier and a brief description of the available spectral libraries.
        """
        self.pfs.downloadToFile(self.baseurl+"/description.txt", os.path.join(self.dpa, "description.txt"), True, self.dlv)
    
    def getREADME(self):
        """
        Get content of README file
        
        Returns
        -------
        README : list of strings
            Content of README file
        """
        rurl = self.baseurl + self.lib + "/README"
        rfn = os.path.join("pyasl", "resBased", "spectrallib", self.lib, "README")
        if not self.pfs.fileExists(rfn):
            self.pfs.downloadToFile(rurl, rfn, True, self.dlv)
        f = self.pfs.requestFile(rfn, 'rt')
        return f.readlines()
    
    def _readDescription(self):
        """
        Read description
        """
        self.description = {}
        ll = self.pfs.requestFile(os.path.join(self.dpa, "description.txt"), 'rt').readlines()
        for l in ll:
            l = l.rstrip("\n")
            s = l.split(":")
            if len(s) == 2:
                self.description[s[0].strip()] = s[1].strip()
    
    def listDescription(self):
        """
        List descriptions of available libraries
        """
        print("-"*80)
        print("--- Description of libraries")
        print("-"*80)
        for k, v in self.description.items():
            print("Key         : ", k)
            print("Description : ", v)
        print("-"*80)

    def refreshInventory(self):
        """
        Re-download inventory.txt file
        """
        self.pfs.downloadToFile(self.invurl, self.invfn, True, self.dlv)        

    def _initLib(self, lib):
        """
        Initialize spectral library
        """
        # Get inventory
        self.invfn = os.path.join("pyasl", "resBased", "spectrallib", lib, "inventory.txt")
        self.invurl = self.baseurl + lib + "/inventory.txt"
        if not self.pfs.fileExists(self.invfn):
            self.pfs.downloadToFile(self.invurl, self.invfn, True, self.dlv)
        # Read inventory
        f = self.pfs.requestFile(self.invfn, 'rt')
        ll = [l.rstrip("\n") for l in f.readlines()]
        self.inv = {}
        for l in ll:
            if l.startswith("#"):
                continue
            s = l.split()
            code = []
            # First is number, last is filename
            for i in range(1,len(s)-1):
                try:
                    x = float(s[i])
                except ValueError:
                    x = s[i]
                code.append(x)
            self.inv[tuple(code)] = s[-1]        
    
    def listInventory(self):
        """
        Print content of inventory
        
        The inventory is stored in the attribute `inv` (dictionary)
        """
        print("-"*80)
        print("--- Inventory")
        print("-"*80)
        for k,v in self.inv.items():
            print(k, v)
        print("-"*80)
    
    def read1dFitsSpec(self, fn):
        """
        Read 1d spectral model
        
        Returns
        -------
        w, f : arrays
            Wavelength [A] and flux
        """
        return readFitsSpec.read1dFitsSpec(fn)
        
    def readMuFits(self, fn):
        """
        Read specific intensity file
        
        Returns
        -------
        
        """
        hl = pyfits.open(fn)
        hkeys = hl[1].header
        w = ((np.arange(hkeys["NAXIS2"]) + 1.0) - hkeys["CRPIX1"]) * hkeys["CDELT1"] + hkeys["CRVAL1"]
        mus = np.array(hl["MUS"].data)
        specs = np.array(hl["I(mu)"].data)
        return w, mus, specs
        
    def requestModel(self, teff, logg, mh=0.0, nex="diskint"):
        """
        Get filename of model spectrum in the library
        
        Model files are downloaded on demand.
        
        Parameters
        ----------
        teff : float
            Effective temperature [K]
        logg : float
            Surface gravity
        mh : float
            Metallicity
        nex : string
            Name extension (e.g., diskint or muspecs)
        
        Returns
        -------
        Spectrum : numpy npz file
            The spectrum in the form of the context of a numpy npz file
        """
        if self.lib == "A":
            p = (float(teff), float(logg), float(mh), nex)
        if not p in self.inv:
            raise(PE.PyAValError("The key '" + str(p) + "' is not in the inventory.", \
                                 where="requestSpectrum", \
                                 solution="Use 'listInventory' to see available combinations."))
        fn = os.path.join(self.dpa, self.lib, self.inv[p])
        if not self.pfs.fileExists(fn):
            surl = self.baseurl + self.lib + "/" + self.inv[p]
            self.pfs.downloadToFile(surl, fn, True, self.dlv)
        fp = self.pfs.getFullPath(fn)
        return fp
    
    def __init__(self, lib="A", refreshDescr=False, dlv=True):
        """
        Access spectral library/ies
        
        Parameters
        ----------
        lib : string, optional
            Library key
        refreshDescr : boolean, optional
            Re-download library description file (default is False)
        dlv : boolean, optional
            Verbose download, default is True
        """
        
        # Download verbose
        self.dlv = dlv
        self.baseurl = "ftp://ftp.hs.uni-hamburg.de/pub/outgoing/czesla/spectrallib/"
        
        self.pfs = pp.pyaFS.PyAFS()
        # Data path
        self.dpa = os.path.join("pyasl", "resBased", "spectrallib")
        self.lib = lib
        
        self.fndescr = os.path.join(self.dpa, "description.txt")
        if (not self.pfs.fileExists(self.fndescr)) or refreshDescr:
            # Get description
            self.refreshDescription()
        self._readDescription()
        
        if not self.lib in self.description:
            raise(PE.PyAValError("No such library key: " + str(self.lib), \
                                 where="SpectralLib", \
                                 solution="Chose one of: " + ','.join(self.description.keys())))
        
        self._initLib(self.lib)
        
        


from __future__ import print_function
import os
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves.configparser as ConfigParser
import six.moves as smo

class PyARC(object):
    """
    """
    
    def __init__(self):
        self.rc = ConfigParser.RawConfigParser()
        self.rc.add_section("pyaRC")
        # By default, keep online access on
        self.goOnline()
    
    def takeOffline(self):
        self.rc.set("pyaRC", "onlineKillSwitch", 1)
    def goOnline(self):
        self.rc.set("pyaRC", "onlineKillSwitch", 0)
    def supposedOnline(self):
        return (not self.rc["pyaRC"]["onlineKillSwitch"])
       
pyaRC = PyARC() 


class PyAConfig(object):
    """
    Provide access to permanent PyA configuration.

    Permanent configuration and data created by/for
    PyAstronomy are stored in a data directory appointed
    to PyAstronomy. The directory name is stored in a file named
    ".pyaConfigWhere", which is located in the user's home directory
    (defined by environment variable `home`). This file contains only
    a single line, which is exactly the name of the PyA data directory.

    In the PyA data directory, this class creates a stub configuration
    file named "pyaConfig.cfg", if the file does not already exist. 

    Attributes
    ----------
    dpath : string
        The name of PyA's data directory (defined by the content of
        the file ".pyaConfigWhere" in the home directory).
        None if no valid directory could be found or created.           .
    configWhere : string
        The full name of the ".pyaConfigWhere" file.
    """

    _rootConfig = None

    def __createConfigStub(self):
        """
        Creates a stub file for configuration if not yet existing.
        """
        if not os.path.isfile(os.path.join(self.dpath, "pyaConfig.cfg")):
            config = ConfigParser.RawConfigParser()
            with open(os.path.join(self.dpath, "pyaConfig.cfg"), 'wt') as configfile:
                config.write(configfile)
        else:
            return

    def __checkDataDir(self):
        """
        Check whether a valid data directory is present.

        This method throws an exception if no valid data directory
        could be found. The check is performed simply by checking
        the `dpath` attribute.
        """
        if self.dpath is None:
            raise(PE.PyAValError("No valid data and configuration directory for PyA could be found.",
                                 solution="Check/Repeat initialization (maybe remove '.pyaConfigWhere' in home directory)."))

    def getDataRoot(self):
        """
        Access the root path to PyA's permanent data and configuration.

        Returns
        -------
        Path : string
            The root path as defined in the file ".pyaConfigWhere" in the
            home directory.
        """
        self.__checkDataDir()
        return self.dpath

    def accessRootConfig(self):
        """
        Provides access to PyA's root configuration file.

        Returns
        -------
        Configuration : ConfigParser object
            The content of the root configuration file "pyaConfig.cfg".
        """
        return PyAConfig._rootConfig

    def _ynQuestion(self, message):
        """
        Ask yes/no question only accepting definite answer

        Parameters
        ----------
        message : string
            The question being asked.

        Returns
        -------
        Answer : boolean
            True if the answer was yes and False if it was no.
        """
        while True:
            yn = smo.input(message)
            if (yn.lower() == "y" or yn.lower() == "yes"):
                return True
            elif (yn.lower() == "n" or yn.lower() == "no"):
                return False
            else:
                print()
                print("  What do you by '" + str(yn) +
                      "'? Please provide a valid input (y/n).")
                print()

    def _locateHomeDirectory(self):
        """ Try to locate HOME directory and return if possible (else return None) """
        hd = None
        hd = os.getenv("HOME")
        if hd is None:
            # Try alternative solution
            hd = os.path.expanduser("~")
        return hd
        
    def __init__(self):
        self.dpath = None
        # Try to locate home directory via environment variable
        self.homeDir = self._locateHomeDirectory()
        if self.homeDir is None:
            PE.warn(PE.PyAValError("Could not find a home directory. Data directory cannot be set up.",
                                   solution="Set 'HOME' environment variable."))
            return

        # Hard-coded name of file, which will only contain path to "real" configuration.
        # This is done, because at one point, information must be stored, which can be
        # found without any extra information.
        self.configWhere = os.path.join(self.homeDir, ".pyaConfigWhere")
        if not os.path.isfile(self.configWhere):
            # Ask user whether she/he wants to configure data path
            print(" -------------- Configure PyA's data path ------------------------")
            print("  Why do you get this message?")
            print("    Most probably PyA tries to save permanent data on your")
            print("    system for the first time. This may be the case when, e.g.,")
            print("    a table is downloaded and saved. These data are stored")
            print("    under PyA's 'data path', i.e., a regular directory on")
            print("    your disk, which PyA can access.")
            print("  Is anything else saved to the disk?")
            print("    The location of the data path")
            print("    directory will be written to the file '.pyaConfigWhere' in")
            print("    your home directory, so that you need not provide it")
            print("    again.")
            print("  Can I delete it?")
            print("    Yes, you are free to delete anything of this at any time.")
            print("  What is a valid data path?")
            print("    You may provide any existing or non-existing path, although")
            print("    it is strongly encouraged to use a fresh directory to avoid any")
            print("    confusion. The given path needs to be absolute, to uniquely")
            print("    identify it.")
            print(" -----------------------------------------------------------------")
            print("")
            if self._ynQuestion("Configure PyA's data path now (y/n)? "):
                print()
                print("  Configure data path now.")
                print()
            else:
                print()
                print("  Configure data path later.")
                print()
                return

            try:
                while True:
                    print(
                        "Please provide a directory where PyA can store data (may already exist):")
                    print("Press enter to accept default; use 'exit' to abort.")
                    suggestion = os.path.join(self.homeDir, "PyAData")
                    dpath = smo.input(
                        "  Path (default = " + suggestion + "): ")
                    if dpath.lower().strip() == "exit":
                        print("  Process aborted. No data path configured.")
                        dpath = None
                        break
                    if dpath == "":
                        dpath = suggestion
                    # Check whether data path is an absolute path
                    if not os.path.isabs(dpath):
                        print("The path you specified (" + dpath +
                              ") is not absolute, but it needs to be.")
                        dpatha = os.path.abspath(dpath)
                        print("Did you intend to use this path:")
                        print("  ", dpatha)
                        if self._ynQuestion("(y/n) ?"):
                            dpath = dpatha
                            break
                        else:
                            # Retry to ask for path
                            continue
                    if not os.path.isdir(dpath):
                        # Try to create the directory
                        os.makedirs(dpath)
                        break
                    else:
                        print("Directory '" + dpath + "' exists.")
                        if self._ynQuestion("Use as PyA data directory (you may want to use, e.g., a subdirectory instead) (y/n) "):
                            if not os.access(dpath, os.W_OK):
                                # Exists and shall be used. Is it writable?
                                print("")
                                print(
                                    "The directory '" + dpath + "' is not writable! Therefore, it cannot be used.")
                                print("")
                                continue
                        break

                if dpath is None:
                    # Process has been aborted
                    print("No valid data path configured.")
                    return

                self.dpath = dpath

                # There is not yet a file '.pyaConfigWhere', which stores the place to look
                # for the real configure-file and data files.
                with open(self.configWhere, 'wt') as f:
                    f.write(self.dpath)

                # All could be created appropriately
                print("PyA data path configured successfully. Using path: ")
                print("  " + self.dpath)

            except Exception as e:
                PE.warn(PE.PyAValError("The directory: '" + str(dpath) + "' " +
                                       " could not be created. Data directory cannot be set up.\n" +
                                       "Error message: " + str(e)))
                os.remove(self.configWhere)
                return

        else:
            # There is a .pyaConfigWhere file.
            try:
                with open(self.configWhere) as cfi:
                    self.dpath = cfi.readline()
            except Exception as e:
                PE.warn(PE.PyAValError("The file " + self.configWhere +
                                       " exists, but could not be opened for reading.",
                                       solution="Check permissions of the file.",
                                       addInfo="Error message: " + str(e)))
                self.dpath = None
            try:
                self.dpath = os.path.realpath(self.dpath)
            except Exception as e:
                PE.warn(PE.PyAValError("Obtained the path '" + self.dpath + "' from .pyaConfigWhere, but" +
                                       "could not expand soft-links etc. (using os.path.realpath).",
                                       solution="Check the path written to .pyaConfigWhere in home directory.",
                                       addInfo="Error message: " + str(e)))
                self.dpath = None
            if not os.path.isdir(self.dpath):
                PE.warn(PE.PyAValError("The directory " + self.dpath + "' " +
                                       "was specified as data path in file '" + self.configWhere + "', " +
                                       "but it does no longer exist! No data can be stored.",
                                       solution=["Delete file '" + self.configWhere + "' to allow reconfiguration.",
                                                 "Modify the file content by giving the (modified) location " +
                                                 "of the data directory."]))
                self.dpath = None
                return
        # Create cfg file unless it already exists
        self.__createConfigStub()
        # Open the "root" configuration file for later access
        if PyAConfig._rootConfig is None:
            PyAConfig._rootConfig = ConfigParser.RawConfigParser()
            PyAConfig._rootConfig.read(
                os.path.join(self.dpath, "pyaConfig.cfg"))

    def saveConfigToFile(self):
        """
        Save current state of configuration to file 'pyaConfig.cfg'.
        """
        with open(os.path.join(self.dpath, "pyaConfig.cfg"), 'wt') as configfile:
            PyAConfig._rootConfig.write(configfile)

    def set(self, section, option, value):
        """
        Add an entry to the global configuration file.

        If the specified section does not already exist, it is created.
        """
        if not PyAConfig._rootConfig.has_section(section):
            PyAConfig._rootConfig.add_section(section)
        PyAConfig._rootConfig.set(section, option, value)
        self.saveConfigToFile()

    def remove_option(self, section, option):
        """
        Remove option
        """
        PyAConfig._rootConfig.remove_option(section, option)
        self.saveConfigToFile()

    def remove_section(self, section):
        """
        Remove section
        """
        PyAConfig._rootConfig.remove_section(section)
        self.saveConfigToFile()






from __future__ import print_function
from __future__ import division
import unittest

class SanityOfSWEETCat(unittest.TestCase):
  
  def setUp(self):
    pass
  
  def tearDown(self):
    pass
  
  def sanity_example(self):
    """
      Checking sanity of SWEETCat example
    """    
    from PyAstronomy import pyasl
    
    sc = pyasl.SWEETCat()
    
    # Access via pandas data frame
    # Show first three rows
    print(sc.data[0:3])
    print()
    
    # Which stars are available?
    print("Available stars: ", sc.data["star"].values)
    print()
    
    # Check whether star is "available"
    star = "XO-5"
    starInTable = (star in sc.data.star.values)
    print("Is " + star + " in the table? ", starInTable)
    
    if starInTable:
      # Get information on that star
      print("Data for star: ", star)
      info = sc.data[sc.data["star"] == star]
      print(info)
      print()
      print("Get the effective temperature")
      teff, erteff = info["teff"], info["erteff"]
      print("Effective temperature of " + star + " = %6.1f +/- %6.1f K" % (teff, erteff))


import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE


def estimateSNR(x, y, xlen, deg=1, controlPlot=False, xlenMode="dataPoints"):
    """
    Estimate Signal to Noise Ratio (SNR) in a data set.

    This function provides a simple algorithm to estimate
    the SNR in a data set. The algorithm subdivides the
    data set into sections with a length determined by the
    `xlen` parameter. Each individual subsection is
    fitted using a polynomial of degree `deg`. The SNR
    is then computed by assuming that the resulting
    chi square value is properly distributed according
    to the chi-square distribution.

    Parameters
    ----------
    x : array
        The abscissa values.
    y : array
        The ordinate values.
    xlen : int or float
        Length of the data subsection considered
        in the computation of the SNR. Whether `xlen`
        refers to the number of data points or a fixed
        subsection of the abscissa is determined by the
        `xlenMode` flag.
    deg : int, optional
        Degree of the polynomial used to fit the
        subsection of the data set. The default is
        one.
    controlPlot : boolean, optional
        If True, a control plot will be shown to
        verify the validity of the estimate.
    xlenMode : string, {"dataPoints", "excerpt", "all"}, optional
        Determines whether `xlen` refers to data points
        or a fixed length scale on the abscissa. If 'all' is specified,
        all available data will be used in the estimation.

    Returns
    -------
    SNR : dictionary
        Contains the SNR estimate for the individual subsections (key 'snrs')
        and the final SNR estimate (mean of all 'sub-SNRs', key 'SNR-Estimate').
    """
    # Stores the SNRs computed from the individual sections.
    snrs = []
    i = 0
    lastdof = None

    if controlPlot:
        try:
            import matplotlib.pylab as plt
        except ImportError:
            raise(PE.PyARequiredImport("Cannot import matplotlib.pylab",
                                       where="estimateSNR",
                                       solution=["Install matplotlib.", "Change `controlPlot` to False."]))

    while True:
        if xlenMode == "dataPoints":
            # xlen is given in units of data points
            indi = range(i * xlen, (i + 1) * xlen)
            i += 1
            if i > len(x) / xlen:
                break
        elif xlenMode == "excerpt":
            # xlen is given in units of the abscissa values
            indi = np.where(np.logical_and(
                x >= i * float(xlen), x < (i + 1) * float(xlen)))[0]
            i += 1
            if i * float(xlen) > max(x):
                break
        elif xlenMode == "all":
            if i > 1:
                break
            indi = range(len(x))
            i += 1
        else:
            raise(PE.PyAValError("Unknown xlenMode (" + str(xlenMode) + ").",
                                 solution="Use either 'dataPoints' or 'excerpt'."))
        # Check whether indi is long enough
        if len(indi) <= (deg + 1):
            # Skip this subsection
            continue

        # Fit polynomial, calculate model and residuals
        poly = np.polyfit(x[indi], y[indi], deg)
        model = np.polyval(poly, x[indi])
        residuals = y[indi] - model

        # The number of "degrees of freedom)
        dof = len(indi) - (deg + 1)
        # Calculate reduced chi square ...
        stdEstimate = np.sqrt((residuals**2).sum() / dof)
        # A brute-force way to compute the expectation of sqrt(1/(X(d)/d)),
        # where X(d) is chi-square distributed with d degrees of freedom.
        if dof != lastdof:
            ccSample = np.random.chisquare(dof, 1000)
            corrFac = np.mean(np.sqrt(1.0 / (ccSample / dof)))
            lastdof = dof
        stdEstimate *= corrFac
        # ... and SNR
        snrs.append(np.mean(model) / stdEstimate)

        if controlPlot:
            if i == 1:
                # This is the first call
                ax1 = plt.subplot(3, 1, 1)
                plt.title("Data (blue) and model (red)")
                ax2 = plt.subplot(3, 1, 2, sharex=ax1)
                plt.title("Residuals")
                ax3 = plt.subplot(3, 1, 3, sharex=ax1)
                plt.title("SNR")
            # Create plot
            plt.subplot(3, 1, 1)
            plt.plot(x[indi], y[indi], 'b.')
            plt.plot(x[indi], model, 'r-')
            plt.subplot(3, 1, 2, sharex=ax1)
            plt.plot(x[indi], residuals, 'g.')
            plt.subplot(3, 1, 3, sharex=ax1)
            plt.plot(np.mean(x[indi]), snrs[-1], 'bp')

    if controlPlot:
        plt.show()

    return {"SNR-Estimate": np.mean(snrs), "snrs": snrs}




def twilightName(sunAlt):
  """
    Determine the type of twilight for a given altitude of the Sun.
    
    Definitions of twilight:
      - Solar altitude > 0    : day
      - Solar altitude <= 0   : civil twilight
      - Solar altitude <= -6  : nautical twilight
      - Solar altitude <= -12 : astronomical twilight
      - Solar altitude <= -18 : night

    Parameters
    ----------
    sunAlt : float
        Altitude of the Sun in degrees.
         
    Returns
    -------
    Twilight flag : string, {"day", "civil twilight", "nautical twilight", "astron. twilight", "night"}
        Specifies the type of twilight corresponding to the
        given solar altitude.
         
        The twilight flag is either "day", "civil twilight", "nautical twilight",
        "astron. twilight", or "night".
  """  
  
  if sunAlt <= -18.0: return "night"
  if sunAlt <= -12.0: return "astron. twilight"
  if sunAlt <=  -6.0: return "nautical twilight"
  if sunAlt <=   0.0: return "civil twilight"
  return "day"